import argparse
import json
import os

import gym
import numpy as np

import configs
from configs import derive_config
from train import run_experiment
from common.const import PATH_TO_BENCHMARKS
from utilities.statistics import increment_mean_var

if __name__ == '__main__':
    all_envs = [e.worker_id for e in list(gym.envs.registry.all())]
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

    parser = argparse.ArgumentParser(description="Perform a comparative benchmarking experiment. Given one task, a "
                                                 "number of agents is trained with different configurations, "
                                                 "for instance comparing different architectures.")
    parser.add_argument("env", type=str, choices=all_envs, help="environment in which the configurations are compared")
    parser.add_argument("name", type=str, nargs="?", default="default",
                        help="the name of the experiment, used for data "
                             "saving, will default to a combination of env "
                             "and configs")
    parser.add_argument("--repetitions", "-r", type=int, help="number of repetitions of each configuration for means"
                        , default=10)
    parser.add_argument("--cycles", "-i", type=int, help="number of cycles during one drill", default=None)
    parser.add_argument("--configs", "-c", type=str, nargs="+", help="a list of configurations to be compared")
    parser.add_argument("--stop-early", "-e", action="store_true", help="allow independent early stopping")
    args = parser.parse_args()

    configurations = {n: derive_config(getattr(configs, n), {"stop_early": args.stop_early}) for n in args.configs}
    results_file_name = f"{args.name}_{args.env}.json"
    results_path = f"{PATH_TO_BENCHMARKS}/{results_file_name}"

    benchmark_dict = {
        "results": {}
    }

    if os.path.exists(results_path):
        with open(results_path, "r") as f:
            benchmark_dict = json.load(f)

    benchmark_dict.update({
        "meta": {
            "reward_threshold": gym.make(args.env).spec.reward_threshold,
        },
    })

    should_init = True
    for conf_name, config in configurations.items():
        config["iterations"] = args.cycles if args.cycles is not None else config["iterations"]
        config["config"] = conf_name
        config["eval"] = True

        conf_up_counter = 1
        original_conf_name = conf_name
        while conf_name in benchmark_dict["results"]:
            conf_compatible = args.cycles == len(benchmark_dict["results"][conf_name]["means"])

            if conf_compatible:
                print(f"Found compatible config; Extending {conf_name}.")
                break

            conf_name = f"{original_conf_name}_{conf_up_counter}"
        else:
            print(f"No compatibles to {original_conf_name} benchmarked, adding new condition {conf_name}.")
            benchmark_dict["results"].update({
                conf_name: {
                    "n": 0
                }
            })

        benchmark_dict["results"][conf_name].update({
            "config": config
        })

        for i in range(args.repetitions):
            print(f"\nRepetition {i + 1}/{args.repetitions} in environment {args.env} with config {conf_name}.")
            reward_history = np.array(run_experiment(
                args.env, config, verbose=False).cycle_reward_history)
            should_init = False

            current_n = benchmark_dict["results"][conf_name]["n"]
            if current_n == 0:
                means, var = reward_history, np.zeros_like(reward_history)
                mean_max, var_max = np.max(reward_history), np.array(0)
            else:
                means, var = increment_mean_var(np.array(benchmark_dict["results"][conf_name]["means"]),
                                                np.array(benchmark_dict["results"][conf_name]["var"]),
                                                reward_history,
                                                np.zeros_like(reward_history),
                                                current_n)
                mean_max, var_max = increment_mean_var(benchmark_dict["results"][conf_name]["mean_max"],
                                                       benchmark_dict["results"][conf_name]["var_max"],
                                                       np.max(reward_history),
                                                       np.array(0),
                                                       current_n)

            benchmark_dict["results"][conf_name].update({
                "n": benchmark_dict["results"][conf_name]["n"] + 1,

                # mean/var per cycle
                "means": means.tolist(),
                "var": var.tolist(),

                # max cycle performance
                "mean_max": mean_max.item(),
                "var_max": var_max.item()
            })

            with open(results_path, "w") as f:
                json.dump(benchmark_dict, f, indent=2)



"""Generation of auxiliary datasets."""

import numpy as np
import tensorflow as tf
from tqdm import tqdm

from angorapy.common.loss import multi_point_euclidean_distance
from angorapy.common.wrappers import make_env
from angorapy.environments import *


def gen_cube_quats_prediction_data(n):
    """Generate dataset of hand images with cubes as data points and the pos and rotation of the cube as targets."""
    hand_env = gym.make("BaseShadowHandEnv-v0")

    sample = hand_env.observation_space.sample()
    X, Y = np.empty((n,) + sample["observation"][0].shape), np.empty((n,) + sample["achieved_goal"].shape)

    hand_env.reset()
    for i in tqdm(range(n), desc="Sampling Data"):
        sample, r, done, info = hand_env.step(hand_env.action_space.sample())
        image = sample["observation"][0] / 255
        quaternion = sample["achieved_goal"]
        X[i] = image
        Y[i] = quaternion

        if done:
            hand_env.reset()

    return tf.convert_to_tensor(X.astype("float32")), tf.convert_to_tensor(Y.astype("float32"))


def gen_hand_pos_prediction_data(n):
    """Generate dataset of hand images with cubes as data points and the pos and rotation of the cube as targets."""
    hand_env = make_env("ReachAbsoluteVisual-v0")

    sample = hand_env.observation_space.sample()
    X, Y = np.empty((n,) + sample["observation"]["vision"].shape), np.empty((n,) + sample["achieved_goal"].shape)

    hand_env.reset()
    for i in tqdm(range(n), desc="Sampling Data"):
        sample, r, done, info = hand_env.step(hand_env.action_space.sample() * 2)  # scale to have extremer actions
        image = sample.vision / 255
        hand_pos = hand_env.get_fingertip_positions()
        X[i], Y[i] = image, hand_pos

        if done:
            hand_env.reset()

    mean_pos = tf.reduce_mean(Y, axis=0)
    print(f"MSE from Mean: {tf.losses.mse(tf.reshape(Y, -1), tf.tile(mean_pos, [Y.shape[0]]))}")
    print(
        f"Euclidean Distance from Mean: {multi_point_euclidean_distance(tf.reshape(Y, -1), tf.tile(mean_pos, [Y.shape[0]]))}")

    # for i in range(20):
    #     index = random.randint(0, X.shape[0])
    #     plt.imshow(X[index, ...])
    #     plt.title(f"Image {index} ({Y[index, ...]})")
    #     plt.show()

    return tf.convert_to_tensor(X.astype("float32")), tf.convert_to_tensor(Y.astype("float32"))


if __name__ == '__main__':
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

    data = gen_cube_quats_prediction_data(1000)



from typing import Dict

import numpy as np


def robot_get_obs(model, data):
    """Returns all joint positions and velocities associated with a robot."""
    joint_names = mj_get_category_names(model, "jnt")
    if data.qpos is not None and joint_names:
        names = [n for n in joint_names if n.startswith(b"robot")]
        return (
            np.array([data.jnt(name).qpos for name in names]).flatten(),
            np.array([data.jnt(name).qvel for name in names]).flatten(),
        )
    return np.zeros(0), np.zeros(0)


def mj_get_category_names(model, category: str):
    # todo make assert
    adresses = getattr(model, f"name_{category}adr")
    return model.names[adresses[0]:].split(b'\x00')[:len(adresses)]


def mj_qpos_dict_to_qpos_vector(model, qpos_dict: Dict):
    """From a given dictionary of joint name-position pairs, create a vector using their order in the model."""
    return np.array(
        [qpos_dict[n] for n in map(lambda x: x.decode("utf-8"), model.names.split(b"\x00")) if n in qpos_dict.keys()]
    )



"""Maker functions that generate model builder from parameters."""
import itertools
from collections import OrderedDict
from functools import partial

from angorapy.models.shadow import build_shadow_brain_base
from angorapy.models.simple import build_simple_models, build_deeper_models, build_wider_models

arguments = OrderedDict(sorted([
    ("model", ["simple", "deeper", "wider", "shadow"]),
    ("model_type", ["ffn", "rnn", "lstm", "gru"]),
    ("blind", [True, False]),
    ("shared", [True, False]),
], key=lambda x: x[0]))

combinations = list(itertools.product(*arguments.values()))
parameter_dicts = list(map(lambda c: dict(zip(arguments.keys(), c)), combinations))

for pd in parameter_dicts:
    if pd["model"] == "simple":
        base_function = build_simple_models
    elif pd["model"] == "shadow":
        base_function = build_shadow_brain_base
    elif pd["model"] == "deeper":
        base_function = build_deeper_models
    elif pd["model"] == "wider":
        base_function = build_wider_models
    else:
        raise ValueError("Unknown model name registered.")

    name_parts = pd.copy()
    name_parts["shared"] = "shared" if name_parts["shared"] else "distinct"
    name_parts["blind"] = "blind" if name_parts["blind"] else ""

    func_name = f"build_{'_'.join([str(val) for n, val in sorted(name_parts.items(), key=lambda x: x[0]) if val])}_models"
    pd.pop("model")
    globals()[func_name] = partial(base_function, **pd)
    globals()[func_name].__name__ = func_name


def get_model_type(model_builder) -> str:
    """Get short name of type of a model builder."""
    for mt in arguments["model_type"]:
        if mt in model_builder.__name__:
            return mt

    return "unknown"


if __name__ == '__main__':
    print("\n".join(f for f in locals() if f.startswith("build_")))


import os
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

import tensorflow as tf
from environments import *

from agent.ppo_agent import PPOAgent
from common.wrappers import make_env
from angorapy.models import get_model_builder

import matplotlib.pyplot as plt

tf.get_logger().setLevel('INFO')

env = make_env("ReachAbsoluteVisual-v0")
agent = PPOAgent(get_model_builder("shadow", "gru"), env, 1024, 8)

state = env.reset()

for i in range(100):
    state, r, dd, info = env.step(env.action_space.sample())

plt.imshow(state.vision / 255)
plt.show()


from abc import ABC
from collections import OrderedDict
from os import path
from typing import Optional

import numpy as np
import mujoco

import gym
from gym import error, logger, spaces

from angorapy.environments.utils import mj_qpos_dict_to_qpos_vector

DEFAULT_SIZE = 480


def convert_observation_to_space(observation):
    if isinstance(observation, dict):
        space = spaces.Dict(
            OrderedDict(
                [
                    (key, convert_observation_to_space(value))
                    for key, value in observation.items()
                ]
            )
        )
    elif isinstance(observation, np.ndarray):
        low = np.full(observation.shape, -float("inf"), dtype=np.float32)
        high = np.full(observation.shape, float("inf"), dtype=np.float32)
        space = spaces.Box(low, high, dtype=observation.dtype)
    else:
        raise NotImplementedError(type(observation), observation)

    return space


class AnthropomorphicEnv(gym.Env, ABC):
    """Superclass for all environments."""

    def __init__(self, model_path, frame_skip, initial_qpos=None):

        if model_path.startswith("/"):
            fullpath = model_path
        else:
            fullpath = path.join(path.dirname(__file__), "assets", model_path)
        if not path.exists(fullpath):
            raise OSError(f"File {fullpath} does not exist")

        self.model = mujoco.MjModel.from_xml_path(filename=fullpath)
        self.data = mujoco.MjData(self.model)
        self._viewers = {}
        self.viewer = None

        self.frame_skip = frame_skip
        self.metadata = {
            "render_modes": ["human", "rgb_array", "depth_array"],
            "render_fps": int(np.round(1.0 / self.dt)),
        }

        if initial_qpos is None or not initial_qpos:
            initial_qpos = self.data.qpos[:]
        else:
            initial_qpos = mj_qpos_dict_to_qpos_vector(self.model, initial_qpos)

        self.initial_state = {
            "qpos": initial_qpos,
            "qvel": self.data.qvel[:]
        }

        self._env_setup(initial_state=self.initial_state)
        self._set_action_space()

        action = self.action_space.sample()
        observation = self._get_obs()
        self._set_observation_space(observation)

    def _set_action_space(self):
        bounds = self.model.actuator_ctrlrange.copy().astype(np.float32)
        low, high = bounds.T
        self.action_space = spaces.Box(low=low, high=high, dtype=np.float32)
        return self.action_space

    def _set_observation_space(self, observation):
        self.observation_space = convert_observation_to_space(observation)
        return self.observation_space

    def _env_setup(self, initial_state):
        raise NotImplementedError

    # methods to override:
    # ----------------------------

    def _get_obs(self):
        raise NotImplementedError

    def reset_model(self):
        """
        Reset the robot degrees of freedom (qpos and qvel).
        Implement this in each subclass.
        """
        raise NotImplementedError

    def viewer_setup(self):
        """
        This method is called when the viewer is initialized.
        Optionally implement this method, if you need to tinker with camera position and so forth.
        """

    # -----------------------------

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ):
        super().reset(seed=seed)

        mujoco.mj_resetData(self.model, self.data)

        ob = self.reset_model()
        if not return_info:
            return ob
        else:
            return ob, {}

    def get_state(self):
        """Get the current state of the simulation."""
        return {
            "qpos": self.data.qpos[:],
            "qvel": self.data.qvel[:],
        }

    def set_state(self, qpos, qvel):
        assert qpos.shape == (self.model.nq,) and qvel.shape == (self.model.nv,), \
            f"State shape [{qpos.shape}|{qvel.shape}] does not fit [{self.model.nq}|{self.model.nv}]"

        self.data.qpos[:] = np.copy(qpos)
        self.data.qvel[:] = np.copy(qvel)
        if self.model.na == 0:
            self.data.act[:] = None
        mujoco.mj_forward(self.model, self.data)

    @property
    def dt(self):
        return self.model.opt.timestep * self.frame_skip

    def do_simulation(self, ctrl, n_frames):
        self._set_action(ctrl)

        for _ in range(n_frames):
            mujoco.mj_step(self.model, self.data)
            self._step_callback()

        mujoco.mj_rnePostConstraint(self.model, self.data)

    def _set_action(self, action):
        if np.array(action).shape != self.action_space.shape:
            raise ValueError("Action dimension mismatch")

        self.data.ctrl[:] = action

    def _step_callback(self):
        pass

    def render(
        self,
        mode="human",
        width=DEFAULT_SIZE,
        height=DEFAULT_SIZE,
        camera_id=None,
        camera_name=None,
    ):
        if mode == "rgb_array" or mode == "depth_array":
            if camera_id is not None and camera_name is not None:
                raise ValueError(
                    "Both `camera_id` and `camera_name` cannot be"
                    " specified at the same time."
                )

            no_camera_specified = camera_name is None and camera_id is None
            if no_camera_specified:
                camera_name = "track"

            if camera_id is None:
                camera_id = mujoco.mj_name2id(
                    self.model,
                    mujoco.mjtObj.mjOBJ_CAMERA,
                    camera_name,
                )

                self._get_viewer(mode).render(width, height, camera_id=camera_id)

        if mode == "rgb_array":
            data = self._get_viewer(mode).read_pixels(width, height, depth=False)
            # original image is upside-down, so flip it
            return data[::-1, :, :]
        elif mode == "depth_array":
            self._get_viewer(mode).render(width, height)
            # Extract depth part of the read_pixels() tuple
            data = self._get_viewer(mode).read_pixels(width, height, depth=True)[1]
            # original image is upside-down, so flip it
            return data[::-1, :]
        elif mode == "human":
            self._get_viewer(mode).render()

    def close(self):
        if self.viewer is not None:
            self.viewer = None
            self._viewers = {}

    def _get_viewer(self, mode, width=DEFAULT_SIZE, height=DEFAULT_SIZE):
        self.viewer = self._viewers.get(mode)
        if self.viewer is None:
            if mode == "human":
                from gym.envs.mujoco.mujoco_rendering import Viewer

                self.viewer = Viewer(self.model, self.data)
            elif mode == "rgb_array" or mode == "depth_array":
                from gym.envs.mujoco.mujoco_rendering import RenderContextOffscreen

                self.viewer = RenderContextOffscreen(
                    width, height, self.model, self.data
                )

            self.viewer_setup()
            self._viewers[mode] = self.viewer
        return self.viewer

    def get_body_com(self, body_name):
        return self.data.body(body_name).xpos

    def state_vector(self):
        return np.concatenate([self.data.qpos.flat, self.data.qvel.flat])



import unittest

import numpy as np

from angorapy.utilities.statistics import increment_mean_var


class StatisticsTest(unittest.TestCase):

    def test_incremental_mean_var(self):
        n_samples = 100000
        sample_dims = 10

        samples = np.array([np.random.randn(1, sample_dims) for _ in range(n_samples)])

        mean, var, n = samples[0], np.zeros((1, sample_dims, )), 1
        for s in samples[1:]:
            s_var = np.zeros((sample_dims,))
            mean, var = increment_mean_var(mean, var, s, s_var, n)
            n += 1

        np_mean = np.mean(samples, axis=0)
        np_var = np.var(samples, axis=0)

        self.assertTrue(np.allclose(mean, np_mean),
                        np.allclose(mean, np_var))


#!/usr/bin/env python
"""Convolutional components/networks."""

import tensorflow as tf


def _residual_block(x, filters, kernel_size=3, stride=1, conv_shortcut=True, name=None):
    """A residual block.

    Arguments:
      x: input tensor.
      filters: integer, filters of the bottleneck layer.
      kernel_size: default 3, kernel size of the bottleneck layer.
      stride: default 1, stride of the first layer.
      conv_shortcut: default True, use convolution shortcut if True,
          otherwise identity shortcut.
      name: string, block label.

    Returns:
      Output tensor for the residual block.
    """
    bn_axis = 3 if tf.keras.backend.image_data_format() == 'channels_last' else 1

    if conv_shortcut:
        # linear convolutional shortcut that matches the sizes
        shortcut = tf.keras.layers.Conv2D(4 * filters, 1, strides=stride, name=name + '_0_conv')(x)
        shortcut = tf.keras.layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5, name=name + '_0_bn')(shortcut)
    else:
        shortcut = x

    x = tf.keras.layers.Conv2D(filters, 1, strides=stride, name=name + '_1_conv')(x)
    x = tf.keras.layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5, name=name + '_1_bn')(x)
    x = tf.keras.layers.Activation('relu', name=name + '_1_relu')(x)

    x = tf.keras.layers.Conv2D(filters, kernel_size, padding='SAME', name=name + '_2_conv')(x)
    x = tf.keras.layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5, name=name + '_2_bn')(x)
    x = tf.keras.layers.Activation('relu', name=name + '_2_relu')(x)

    x = tf.keras.layers.Conv2D(4 * filters, 1, name=name + '_3_conv')(x)
    x = tf.keras.layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5, name=name + '_3_bn')(x)

    x = tf.keras.layers.Add(name=name + '_add')([shortcut, x])
    x = tf.keras.layers.Activation('relu', name=name + '_out')(x)

    return x


# VISUAL ENCODING

def _build_openai_encoder(shape, out_shape, name="visual_component", batch_size=None):
    """Shallow AlexNet Version. Original number of channels are too large for normal GPU memory.

    Args:
        out_shape:
    """
    inputs = tf.keras.Input(batch_shape=(batch_size,) + tuple(shape))

    # first layer
    x = tf.keras.layers.Conv2D(32, 5, 1, padding="valid")(inputs)  # 196 x 196
    x = tf.keras.layers.Activation("relu")(x)

    # second layer
    x = tf.keras.layers.Conv2D(32, 3, 1, padding="valid")(x)  # 195 x 195
    x = tf.keras.layers.Activation("relu")(x)

    # pooling
    x = tf.keras.layers.MaxPool2D(3, 3)(x)  # 64 x 64

    # resnet
    x = _residual_block(x, 16, 3, 3, name="res1")
    x = _residual_block(x, 32, 3, 3, name="res2")
    x = _residual_block(x, 64, 3, 3, name="res3")
    x = _residual_block(x, 64, 3, 3, name="res4")

    # fully connected
    x = tf.keras.layers.Activation("softmax")(x)
    x = tf.keras.layers.Flatten()(x)
    x = tf.keras.layers.Dense(128)(x)
    x = tf.keras.layers.Activation("relu")(x)

    # output
    x = tf.keras.layers.Dense(out_shape)(x)

    return tf.keras.Model(inputs=inputs, outputs=x, name=name)


def _build_openai_small_encoder(shape, out_shape, name="visual_component", batch_size=None):
    """Shallow AlexNet Version. Original number of channels are too large for normal GPU memory.

    Args:
        out_shape:
    """
    inputs = tf.keras.Input(shape=shape, batch_size=batch_size, name=name + "_input")

    # first layer
    x = tf.keras.layers.Conv2D(32, 5, 1, padding="valid")(inputs)  # 96 x 96
    x = tf.keras.layers.Activation("relu")(x)

    # second layer
    x = tf.keras.layers.Conv2D(32, 3, 1, padding="valid")(x)  # 94 x 94
    x = tf.keras.layers.Activation("relu")(x)

    # pooling
    x = tf.keras.layers.MaxPool2D(3, 3)(x)  # 31 x 31

    # resnet
    x = _residual_block(x, 16, 3, 3, name="res1")
    x = _residual_block(x, 32, 3, 3, name="res2")
    x = _residual_block(x, 64, 3, 3, name="res3")

    # fully connected
    x = tf.keras.layers.Activation("softmax")(x)
    x = tf.keras.layers.Flatten()(x)
    x = tf.keras.layers.Dense(128, kernel_regularizer=tf.keras.regularizers.L2(l2=0.001))(x)
    x = tf.keras.layers.Activation("relu")(x)

    # output
    x = tf.keras.layers.Dense(out_shape)(x)

    return tf.keras.Model(inputs=inputs, outputs=x, name=name)


def _build_alexnet_encoder(shape, batch_size=None, name="visual_component"):
    """Shallow AlexNet Version. Original number of channels are too large for normal GPU memory."""
    inputs = tf.keras.Input(batch_shape=(batch_size,) + tuple(shape))

    # first layer
    x = tf.keras.layers.Conv2D(32, 11, 4)(inputs)
    x = tf.keras.layers.Activation("tanh")(x)
    x = tf.keras.layers.MaxPool2D(3, 2)(x)

    # second layer
    x = tf.keras.layers.Conv2D(32, 5, 1, padding="same")(x)
    x = tf.keras.layers.Activation("tanh")(x)
    x = tf.keras.layers.MaxPool2D(3, 2)(x)

    # third layer
    x = tf.keras.layers.Conv2D(64, 3, 1, padding="same")(x)
    x = tf.keras.layers.Activation("tanh")(x)

    # fourth layer
    x = tf.keras.layers.Conv2D(64, 3, 1, padding="same")(x)
    x = tf.keras.layers.Activation("tanh")(x)

    # fifth layer
    x = tf.keras.layers.Conv2D(32, 3, 1, padding="same")(x)
    x = tf.keras.layers.Activation("tanh")(x)

    x = tf.keras.layers.MaxPool2D(3, 2)(x)

    # fully connected layers
    x = tf.keras.layers.Flatten()(x)
    x = tf.keras.layers.Dense(512)(x)
    x = tf.keras.layers.Activation("tanh")(x)
    x = tf.keras.layers.Dense(512)(x)
    x = tf.keras.layers.Activation("tanh")(x)

    return tf.keras.Model(inputs=inputs, outputs=x, name=name)


# VISUAL DECODING

def _build_visual_decoder(input_dim):
    model = tf.keras.Sequential((
        tf.keras.layers.Dense(512, input_dim=input_dim, activation="relu"),
        tf.keras.layers.Dense(512, activation="relu"),
        tf.keras.layers.Dense(1152, activation="relu"),

        tf.keras.layers.Reshape((6, 6, 32)),
        tf.keras.layers.Conv2DTranspose(32, 3, 2),
        tf.keras.layers.Conv2DTranspose(64, 3, 1, activation="relu", padding="same"),
        tf.keras.layers.Conv2DTranspose(64, 3, 1, activation="relu", padding="same"),
        tf.keras.layers.Conv2DTranspose(32, 3, 2),
        tf.keras.layers.Conv2DTranspose(32, 5, 1, activation="relu", padding="same"),
        tf.keras.layers.Conv2DTranspose(32, 3, 2),
        tf.keras.layers.Conv2DTranspose(3, 11, 4, activation="sigmoid"),
    ))

    model.build()
    return model


if __name__ == "__main__":
    conv_comp = _build_openai_small_encoder((100, 100, 3), 7)
    print(conv_comp(tf.random.normal((1, 100, 100, 3))).shape)
    conv_comp.summary()
    #
    # print("\n\n\n")
    # conv_comp = _build_openai_encoder((200, 200, 3), 7)
    # conv_comp.summary()



#!/usr/bin/env python
"""BaseShadowHandEnv Environment Wrappers."""
import abc
import copy
import os
import random
from typing import Callable, Union

import gym
import numpy as np
from gym import spaces
from gym.utils import seeding
from angorapy.common import reward

from angorapy.common.const import N_SUBSTEPS
from angorapy.configs.reward_config import resolve_config_name
from angorapy.environments.mujoco_env import AnthropomorphicEnv
from angorapy.environments.utils import mj_get_category_names

FINGERTIP_SITE_NAMES = [
    'robot0:S_fftip',
    'robot0:S_mftip',
    'robot0:S_rftip',
    'robot0:S_lftip',
    'robot0:S_thtip',
]

DEFAULT_INITIAL_QPOS = {
    'robot0:WRJ1': -0.16514339750464327,
    'robot0:WRJ0': -0.31973286565062153,
    'robot0:FFJ3': 0.14340512546557435,
    'robot0:FFJ2': 0.32028208333591573,
    'robot0:FFJ1': 0.7126053607727917,
    'robot0:FFJ0': 0.6705281001412586,
    'robot0:MFJ3': 0.000246444303701037,
    'robot0:MFJ2': 0.3152655251085491,
    'robot0:MFJ1': 0.7659800313729842,
    'robot0:MFJ0': 0.7323156897425923,
    'robot0:RFJ3': 0.00038520700007378114,
    'robot0:RFJ2': 0.36743546201985233,
    'robot0:RFJ1': 0.7119514095008576,
    'robot0:RFJ0': 0.6699446327514138,
    'robot0:LFJ4': 0.0525442258033891,
    'robot0:LFJ3': -0.13615534724474673,
    'robot0:LFJ2': 0.39872030433433003,
    'robot0:LFJ1': 0.7415570009679252,
    'robot0:LFJ0': 0.704096378652974,
    'robot0:THJ4': 0.003673823825070126,
    'robot0:THJ3': 0.5506291436028695,
    'robot0:THJ2': -0.014515151997119306,
    'robot0:THJ1': -0.0015229223564485414,
    'robot0:THJ0': -0.7894883021600622,
}

MODEL_PATH = os.path.join(os.path.dirname(__file__), 'assets/hand/', 'shadowhand.xml')
# MODEL_PATH_MANIPULATE = os.path.join(os.path.dirname(__file__), 'assets/hand/', 'shadowhand.xml')
MODEL_PATH_MANIPULATE = os.path.join(os.path.dirname(__file__), 'assets/hand/', 'shadowhand_manipulate.xml')


def generate_random_sim_qpos(base: dict) -> dict:
    """Generate a random state of the simulation."""
    for key, val in base.items():
        if key in ["robot0:WRJ1", "robot0:WRJ0"]:
            continue  # do not randomize the wrist

        base[key] = val + val * random.gauss(0, 0.1)

    return base


def get_palm_position(model):
    """Return the robotic hand's palm's center."""
    return model.body("robot0:palm").pos


def get_fingertip_distance(ft_a, ft_b):
    """Return the distance between two vectors representing finger tip positions."""
    assert ft_a.shape == ft_b.shape
    return np.linalg.norm(ft_a - ft_b, axis=-1)


class BaseShadowHandEnv(AnthropomorphicEnv):  #, abc.ABC):
    """Base class for all shadow hand environments, setting up mostly visual characteristics of the environment."""

    continuous = True
    discrete_bin_count = 11

    def __init__(self,
                 initial_qpos,
                 distance_threshold,
                 n_substeps=N_SUBSTEPS,
                 delta_t=0.002,
                 relative_control=True,
                 model=MODEL_PATH):
        gym.utils.EzPickle.__init__(**locals())

        self.relative_control = relative_control
        self._touch_sensor_id_site_id = []
        self._touch_sensor_id = []
        self.distance_threshold = distance_threshold
        self.reward_type = "dense"
        self._freeze_wrist = False

        # time control
        self._delta_t_control: float = delta_t
        self._delta_t_simulation: float = delta_t
        self._simulation_steps_per_control_step: int = int(self._delta_t_control // self._delta_t_simulation)
        self._always_render_mode = False

        super(BaseShadowHandEnv, self).__init__(model_path=model, frame_skip=n_substeps, initial_qpos=initial_qpos)

        self.model.opt.timestep = delta_t
        self.original_n_substeps = n_substeps

        self.seed()
        self.initial_state = copy.deepcopy(self.get_state())

        obs = self._get_obs()

    def _set_action_space(self):
        if self.continuous:
            self.action_space = spaces.Box(-1., 1., shape=(20,), dtype=float)
        else:
            self.action_space = spaces.MultiDiscrete(np.ones(20) * BaseShadowHandEnv.discrete_bin_count)
            self.discrete_action_values = np.linspace(-1, 1, BaseShadowHandEnv.discrete_bin_count)

    def _set_observation_space(self, obs):
        # bounds are set to max of dtype to avoid infinity warnings
        self.observation_space = spaces.Dict(dict(
            desired_goal=spaces.Box(np.finfo(np.float32).min, np.finfo(np.float32).max, shape=obs['achieved_goal'].shape, dtype=float),
            achieved_goal=spaces.Box(np.finfo(np.float32).min, np.finfo(np.float32).max, shape=obs['achieved_goal'].shape, dtype=float),
            observation=spaces.Dict(
                {name: spaces.Box(np.finfo(np.float32).min, np.finfo(np.float32).max, shape=val.shape, dtype=float)
                 for name, val in obs['observation'].dict().items()}
            ),
        ))

    def set_delta_t_simulation(self, new: float):
        """Set new value for the simulation delta t."""
        assert np.isclose(self._delta_t_control % new, 0, rtol=1.e-3, atol=1.e-4), \
            f"Delta t of simulation must divide control delta t into integer " \
            f"parts, but gives {self._delta_t_control % new}."

        self._delta_t_simulation = new
        self.model.opt.timestep = self._delta_t_simulation

        self._simulation_steps_per_control_step = int(self._delta_t_control // self._delta_t_simulation) * self.original_n_substeps

    def toggle_wrist_freezing(self):
        """Toggle flag preventing the wrist from moving."""
        self._freeze_wrist = not self._freeze_wrist
        print("Wrist movements are now frozen.")

    def compute_reward(self, achieved_goal, goal, info):
        """Compute reward with additional success bonus."""
        return self.reward_function(self, achieved_goal, goal, info)

    @abc.abstractmethod
    def assert_reward_setup(self):
        pass

    def set_reward_function(self, function: Union[str, Callable]):
        """Set the environment reward function by its config identifier or a callable."""
        if isinstance(function, str):
            try:
                function = getattr(reward, function.split(".")[0])
            except AttributeError:
                raise AttributeError("Reward function unknown.")

        self.reward_function = function

    def set_reward_config(self, new_config: Union[str, dict]):
        """Set the environment's reward configuration by its identifier or a dict."""
        if isinstance(new_config, str):
            new_config: dict = resolve_config_name(new_config)

        self.reward_config = new_config
        if "SUCCESS_DISTANCE" in self.reward_config.keys():
            self.distance_threshold = self.reward_config["SUCCESS_DISTANCE"]

        self.assert_reward_setup()

    # INFORMATION METHODS
    def get_fingertip_positions(self):
        """Get positions of all fingertips in euclidean space. Each position is encoded by three floating point numbers,
        as such the output is a 15-D numpy array."""
        goal = [self.data.get_site_xpos(name) for name in FINGERTIP_SITE_NAMES]
        return np.array(goal).flatten()

    # ENV METHODS

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def step(self, action: np.ndarray):
        if self.continuous:
            action = np.clip(action, self.action_space.low, self.action_space.high)
        else:
            action = self.discrete_action_values[action.astype(int)]

        if self._freeze_wrist:
            action[:2] = 0

        # perform simulation
        self.do_simulation(action, n_frames=self.original_n_substeps)

        # read out observation from simulation
        obs = self._get_obs()

        done = False
        info = {
            'is_success': self._is_success(obs['achieved_goal'], self.goal),
        }

        reward = self.compute_reward(obs['achieved_goal'], self.goal, info)

        return obs, reward, done, info

    def reset(self, return_info=False, **kwargs):
        # Attempt to reset the simulator. Since we randomize initial conditions, it
        # is possible to get into a state with numerical issues (e.g. due to penetration or
        # Gimbel lock) or we may not achieve an initial condition (e.g. an object is within the hand).
        # In this case, we just keep randomizing until we eventually achieve a valid initial
        # configuration.
        super(BaseShadowHandEnv, self).reset(**kwargs)
        did_reset_sim = False

        while not did_reset_sim:
            did_reset_sim = self._reset_sim()
        self.goal = self._sample_goal().copy()

        obs = self._get_obs()

        if return_info:
            return obs, {}
        else:
            return obs

    def reset_model(self):
        self.set_state(**self.initial_state)

    def close(self):
        if self.viewer is not None:
            # self.viewer.finish()
            self.viewer = None
            self._viewers = {}

    def render(self, mode='human', width=500, height=500):
        self._render_callback(render_targets=True)
        if mode == 'rgb_array':
            self._get_viewer(mode).render(width, height)
            # window size used for old mujoco-py:
            data = self._get_viewer(mode).read_pixels(width, height, depth=False)
            # original image is upside-down, so flip it
            return data[::-1, :, :]
        elif mode == 'human':
            self._get_viewer(mode).render()

    # Extension methods
    # ----------------------------

    def _reset_sim(self):
        """Resets a simulation and indicates whether or not it was successful.
        If a reset was unsuccessful (e.g. if a randomized state caused an error in the
        simulation), this method should indicate such a failure by returning False.
        In such a case, this method will be called again to attempt a the reset again.
        """
        self.reset_model()

        return True

    @abc.abstractmethod
    def _get_obs(self):
        pass

    def _set_action(self, action):
        assert action.shape == (20,)

        actuator_names = mj_get_category_names(self.model, "actuator")
        ctrlrange = self.model.actuator_ctrlrange
        actuation_range = (ctrlrange[:, 1] - ctrlrange[:, 0]) / 2.
        if self.relative_control:
            actuation_center = np.zeros_like(action)
            for i in range(self.data.ctrl.shape[0]):
                actuation_center[i] = self.data.jnt(actuator_names[i].replace(b':A_', b':')).qpos
            for joint_name in ['FF', 'MF', 'RF', 'LF']:
                act_idx = actuator_names.index(str.encode(f'robot0:A_{joint_name}J1'))
                actuation_center[act_idx] += self.data.jnt(str.encode(f'robot0:{joint_name}J0')).qpos
        else:
            actuation_center = (ctrlrange[:, 1] + ctrlrange[:, 0]) / 2.

        self.data.ctrl[:] = actuation_center + action * actuation_range
        self.data.ctrl[:] = np.clip(self.data.ctrl, ctrlrange[:, 0], ctrlrange[:, 1])

    @abc.abstractmethod
    def _is_success(self, achieved_goal, desired_goal):
        pass

    @abc.abstractmethod
    def _sample_goal(self):
        pass

    def _env_setup(self, initial_state):
        """Initial configuration of the environment. Can be used to configure initial state
        and extract information from the simulation."""
        for k, v in zip(mj_get_category_names(self.model, "sensor"), self.model.sensor_adr):
            if b'robot0:TS_' in k:
                # self._touch_sensor_id_site_id.append((v, self.model._site_name2id[k.replace('robot0:TS_', 'robot0:T_')]))
                self._touch_sensor_id.append(v)

        if initial_state is not None:
            self.set_state(**initial_state)

    @abc.abstractmethod
    def _render_callback(self, render_targets=False):
        """A custom callback that is called before rendering. Can be used to implement custom visualizations."""
        pass

    def _step_callback(self):
        """A custom callback that is called after stepping the simulation. Can be used
        to enforce additional constraints on the simulation state.
        """
        pass

    def _viewer_setup(self):
        body_id = self.model.body_name2id('robot0:palm')
        lookat = self.data.body_xpos[body_id]

        for idx, value in enumerate(lookat):
            self.viewer.cam.lookat[idx] = value

        # hand color
        self.model.mat_rgba[2] = np.array([29, 33, 36, 255]) / 255  # hand
        # self.sim.model.mat_rgba[2] = np.array([200, 200, 200, 255]) / 255  # hand

        # background color
        self.model.mat_rgba[4] = np.array([255, 255, 255, 255]) / 255  # background
        # self.sim.model.mat_rgba[4] = np.array([159, 41, 54, 255]) / 255  # background
        # self.sim.model.geom_rgba[48] = np.array([0.5, 0.5, 0.5, 0])

        self.viewpoint = "topdown"

        if self.viewpoint == "topdown":
            # rotate camera to top down view
            self.viewer.cam.distance = 0.4  # zoom in
            self.viewer.cam.azimuth = -90.0  # wrist to the bottom
            self.viewer.cam.elevation = -90.0  # top down view
            self.viewer.cam.lookat[1] -= 0.07  # slightly move forward
        elif self.viewpoint == "side":
            # rotate camera to side view
            self.viewer.cam.distance = 0.35  # zoom in
            self.viewer.cam.azimuth = 25.0  # top down view
            self.viewer.cam.elevation = -45.0  # top down view
            self.viewer.cam.lookat[1] -= 0.04  # slightly move forward
        else:
            raise NotImplementedError("Unknown Viewpoint.")


if __name__ == '__main__':
    hand = BaseShadowHandEnv(
        initial_qpos=DEFAULT_INITIAL_QPOS,
        distance_threshold=0.2
    )



from inspect import getfullargspec as fargs
from typing import List, Union

import numpy
import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import TimeDistributed

from angorapy.utilities.util import flatten


def is_recurrent_model(model: tf.keras.Model):
    """Check if given model is recurrent (i.e. contains a recurrent layer of any sort)"""
    for layer in extract_layers(model):
        if isinstance(layer, tf.keras.layers.RNN):
            return True

    return False


def requires_batch_size(model_builder) -> bool:
    """Check if model building function requires a batch size when building."""
    return "bs" in fargs(model_builder).args + fargs(model_builder).kwonlyargs


def requires_sequence_length(model_builder) -> bool:
    """Check if model building function requires a sequenc_length when building."""
    return "sequence_length" in fargs(model_builder).args + fargs(model_builder).kwonlyargs


def list_layer_names(network, only_para_layers=True) -> List[str]:
    """Get a list of unique string representations of all layers in the network."""
    if only_para_layers:
        return [layer.name for layer in extract_layers(network) if
                not isinstance(layer, (tf.keras.layers.Activation, tf.keras.layers.InputLayer))]
    else:
        return [layer.name for layer in extract_layers(network) if
                not isinstance(layer, tf.keras.layers.InputLayer)]


def get_layer_names(model: tf.keras.Model):
    """Get names of all (outer) layers in the model."""
    return [layer.name if not isinstance(layer, TimeDistributed) else layer.layer.name for layer in model.layers]


def extract_layers(network: tf.keras.Model, unfold_tds: bool = False) -> List[tf.keras.layers.Layer]:
    """Recursively extract layers from a potentially nested list of Sequentials of unknown depth."""
    if not hasattr(network, "layers"):
        return [network]

    layers = []
    for l in network.layers:
        if isinstance(l, tf.keras.Model) or isinstance(l, tf.keras.Sequential):
            layers.append(extract_layers(l))
        elif isinstance(l, tf.keras.layers.TimeDistributed) and unfold_tds:
            if isinstance(l.layer, tf.keras.Model) or isinstance(l.layer, tf.keras.Sequential):
                layers.append(extract_layers(l.layer))
            else:
                layers.append(l.layer)
        else:
            layers.append(l)

    return flatten(layers)


def get_layers_by_names(network: tf.keras.Model, layer_names: List[str]):
    """Get a list of layers identified by their names from a network."""
    layers = extract_layers(network) + network.layers
    all_layer_names = [l.name for l in layers]

    assert all(ln in all_layer_names for ln in layer_names), "Cannot find layer name in network extraction."

    return [layers[all_layer_names.index(layer_name)] for layer_name in layer_names]


def build_sub_model_to(network: tf.keras.Model, tos: Union[List[str], List[tf.keras.Model]], include_original=False):
    """Build a sub model of a given network that has (multiple) outputs at layer activations defined by a list of layer
    names."""
    layers = get_layers_by_names(network, tos) if isinstance(tos[0], str) else tos
    outputs = []

    # probe layers to check if model can be build to them
    for layer in layers:
        success = False
        layer_input_id = 0
        while not success:
            success = True
            try:
                tf.keras.Model(inputs=[network.input], outputs=[layer.get_output_at(layer_input_id)])
            except ValueError as ve:
                if len(ve.args) > 0 and ve.args[0].split(" ")[0] == "Graph":
                    layer_input_id += 1
                    success = False
                else:
                    raise ValueError(f"Cannot use layer {layer.name}. Error: {ve.args}")
            else:
                outputs.append([layer.get_output_at(layer_input_id)])

    if include_original:
        outputs = outputs + network.outputs

    return tf.keras.Model(inputs=[network.input], outputs=outputs)


def build_sub_model_from(network: tf.keras.Model, from_layer_name: str):
    """EXPERIMENTAL: NOT GUARANTEED TO WORK WITH ALL ARCHITECTURES.

    Build a sub model of a given network that has inputs at layers defined by a given name.
    Outputs will remain the network outputs."""
    all_layer_names = [l.name for l in network.layers]

    first_layer_index = all_layer_names.index(from_layer_name)
    first_layer = network.layers[first_layer_index]

    new_input = tf.keras.layers.Input(first_layer.input_shape[1:])
    x = first_layer(new_input)
    for layer in network.layers[first_layer_index + 1:]:
        x = layer(x)

    return tf.keras.Model(inputs=new_input, outputs=x)


def get_component(model: tf.keras.Model, name: str):
    """Get outer layer/component by name."""
    for layer in model.layers:
        layer_name = layer.name
        if isinstance(layer, TimeDistributed):
            layer_name = layer.layer.name

        if layer_name == name:
            return layer


def reset_states_masked(model: tf.keras.Model, mask: List):
    """Reset a stateful model'serialization states only at the samples in the batch that are specified by the mask.

    The mask should be a list of length 'batch size' and contain one at every position where the state should be reset,
    and zeros otherwise (booleans possible too)."""

    # extract recurrent layers by their superclass RNN
    recurrent_layers = [layer for layer in extract_layers(model) if isinstance(layer, tf.keras.layers.RNN)]

    for layer in recurrent_layers:
        current_states = [state.numpy() for state in layer.states]
        initial_states = 0
        new_states = []
        for current_state in current_states:
            expanded_mask = numpy.tile(numpy.rot90(numpy.expand_dims(mask, axis=0)), (1, current_state.shape[-1]))
            masked_reset_state = np.where(expanded_mask, initial_states, current_state)
            new_states.append(masked_reset_state)

        layer.reset_states(new_states)


def calc_max_memory_usage(model: tf.keras.Model):
    """Calculate memory requirement of a model per sample in bits."""
    layers = extract_layers(model)
    n_shapes = int(numpy.sum(
        [numpy.prod(numpy.array([s if isinstance(s, int) else 1 for s in l.output_shape])) for l in layers]))
    n_parameters = model.count_params()

    # memory needed for saving activations during gradient calculation
    n_activations = 0
    for l in layers:
        if len(l.trainable_variables) == 0 or l.output_shape is None:
            continue

        activation_shapes = l.output_shape
        if not isinstance(activation_shapes[0], tuple):
            activation_shapes = [tuple(activation_shapes)]

        print(activation_shapes)

    print(n_activations)

    total_memory = (n_shapes + n_parameters + n_activations) * 32

    return total_memory * 1.1641532182693481 * 10 ** -10


CONVOLUTION_BASE_CLASS = tf.keras.layers.Conv2D.__bases__[0]


def is_conv(layer):
    """Check if layer is convolutional."""
    return isinstance(layer, CONVOLUTION_BASE_CLASS)



#!/usr/bin/env python
"""Evaluate a loaded agent on a task."""
import argparse
import os
import time

import gym

from agent.ppo_agent import PPOAgent
from analysis.investigation import Investigator
from common.const import BASE_SAVE_PATH, PATH_TO_EXPERIMENTS
from common.wrappers import make_env
import tensorflow as tf

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

hrda_id = 1623868261
ra_id = 1623777689
rant_id = 1623748917

id = rant_id

start = time.time()
agent = PPOAgent.from_agent_state(id, "best")
print(f"Agent {id} successfully loaded.")

investigator = Investigator.from_agent(agent)
env = agent.env
env = make_env("NRP" + env.unwrapped.spec.id)

print(f"Evaluating on {env.unwrapped.spec.id} with {env.unwrapped.sim.nsubsteps} substeps.")

for i in range(100):
    investigator.render_episode(env, slow_down=False)



import json
import os
import re
import shutil
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from angorapy.utilities.monitor.statistical_plots import plot_episode_box_plots, plot_per_receptor_mean

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
os.environ['CUDA_VISIBLE_DEVICES'] = '-1'

import flask
from flask import request, Blueprint, send_from_directory
from flask_jsglue import JSGlue

from angorapy.agent.ppo_agent import PPOAgent
from angorapy.common.const import PATH_TO_EXPERIMENTS, BASE_SAVE_PATH
from angorapy.utilities.monitor.training_plots import plot_memory_usage, plot_execution_times, plot_preprocessor, \
    plot_reward_progress, plot_loss, plot_length_progress, plot_distribution, compare_reward_progress, \
    grouped_reward_progress
from angorapy.utilities.statistics import ignore_none

from angorapy.monitor import app


agents = Blueprint("agents", __name__, static_folder="storage/saved_models/states")
exps = Blueprint("exps", __name__, static_folder="storage/experiments")

app.register_blueprint(agents)
app.register_blueprint(exps)

jsglue = JSGlue(app)


@app.route("/")
def overview():
    """Write Overview page."""
    exp_dir = PATH_TO_EXPERIMENTS
    experiment_paths = [os.path.join(exp_dir, p) for p in os.listdir(exp_dir)]

    experiments = {}
    envs_available = set()
    for exp_path in experiment_paths:

        eid_m = re.match("[0-9]+", str(exp_path.split("/")[-1]))
        if eid_m:
            eid = eid_m.group(0)
            model_path = os.path.join(BASE_SAVE_PATH, eid)

            if os.path.isfile(os.path.join(exp_path, "progress.json")):
                with open(os.path.join(exp_path, "progress.json"), "r") as f:
                    progress = json.load(f)

                with open(os.path.join(exp_path, "meta.json"), "r") as f:
                    meta = json.load(f)

                model_available = False
                if os.path.isfile(os.path.join(model_path, "best/weights.index")):
                    model_available = True

                reward_threshold = None if meta["environment"]["reward_threshold"] == "None" else float(
                    meta["environment"]["reward_threshold"])
                iterations = len(progress["rewards"]["mean"])
                max_performance = ignore_none(max, progress["rewards"]["mean"])
                is_success = False
                if iterations >= 0 and reward_threshold is None:
                    is_success = "maybe"
                elif max_performance is not None and max_performance > reward_threshold:
                    is_success = True

                architecture = "Any" if "architecture" not in meta["hyperparameters"] else meta["hyperparameters"]["architecture"]
                model = "Any" if "model" not in meta["hyperparameters"] else meta["hyperparameters"]["model"]

                experiments.update({
                    eid: {
                        "env": meta["environment"]["name"],
                        "date": meta["date"],
                        "host": meta["host"] if "host" in meta else "unknown",
                        "iterations": iterations,
                        "max_reward": max_performance if iterations > 0 else "N/A",
                        "is_success": is_success,
                        "bookmark": meta["bookmark"] if "bookmark" in meta else False,
                        "config_name": meta["config"] if "config" in meta else "unknown",
                        "reward": meta["reward_function"] if "reward_function" in meta else "None",
                        "model": architecture + "[" + model + "]",
                        "model_available": model_available
                    }
                })

                envs_available.add(meta["environment"]["name"])

    return flask.render_template("overview.html", exps=experiments, envs_available=envs_available)


@app.route("/benchmarks")
def benchmarks():
    """Write Benchmark page."""
    exp_dir = PATH_TO_EXPERIMENTS
    experiment_paths = [os.path.join(exp_dir, p) for p in os.listdir(exp_dir)]

    experiments = {}
    envs_available = set()
    for path in experiment_paths:
        eid_m = re.match("[0-9]+", str(path.split("/")[-1]))
        if eid_m:
            eid = eid_m.group(0)

            if os.path.isfile(os.path.join(path, "progress.json")):
                with open(os.path.join(path, "progress.json"), "r") as f:
                    progress = json.load(f)

                with open(os.path.join(path, "meta.json"), "r") as f:
                    meta = json.load(f)

                # for vn, vector in progress.items():
                #     replace()

                reward_threshold = None if meta["environment"]["reward_threshold"] == "None" else float(
                    meta["environment"]["reward_threshold"])
                iterations = len(progress["rewards"]["mean"])
                experiments.update({
                    eid: {
                        "env": meta["environment"]["name"],
                        "date": meta["date"],
                        "host": meta["host"] if "host" in meta else "unknown",
                        "iterations": iterations,
                        "max_reward": max(progress["rewards"]["mean"]) if iterations > 0 else "N/A",
                        "is_success": False if iterations == 0 else ("maybe" if reward_threshold is None else max(
                            progress["rewards"]["mean"]) > reward_threshold),
                        "bookmark": meta["bookmark"] if "bookmark" in meta else False
                    }
                })

                envs_available.add(meta["environment"]["name"])

    return flask.render_template("overview.html", exps=experiments, envs_available=envs_available)


@app.route("/bookmark", methods=("POST", "GET"))
def bookmark():
    """Bookmark an experiment."""
    if request.method == "POST":
        try:
            exp_id = int(request.json['id'])
            path = f"{PATH_TO_EXPERIMENTS}{exp_id}"

            with open(os.path.join(path, "meta.json"), "r") as f:
                meta = json.load(f)
                current_status = "bookmark" in meta and meta["bookmark"]

            meta.update({"bookmark": not current_status})

            with open(os.path.join(path, "meta.json"), "w") as f:
                json.dump(meta, f)

        except Exception as e:
            return {"success": e.__repr__()}
    else:
        return {"success": "no post"}

    return {"success": "success"}


@app.route("/experiment/<int:exp_id>", methods=("POST", "GET"))
def show_experiment(exp_id):
    """Show experiment of given ID."""
    experiment_paths = sorted([int(p) for p in os.listdir(f"{PATH_TO_EXPERIMENTS}") if re.match("^[0-9]*$", p)])
    current_index = experiment_paths.index(exp_id)

    path = f"{PATH_TO_EXPERIMENTS}/{exp_id}"
    with open(os.path.join(path, "progress.json"), "r") as f:
        progress = json.load(f)

    with open(os.path.join(path, "meta.json"), "r") as f:
        meta = json.load(f)

    info = dict(
        env=meta["environment"]["name"],
        config=meta["config"] if "config" in meta else None,
        host=meta["host"] if "host" in meta else "unknown",
        current_id=exp_id,
        next_id=experiment_paths[current_index + 1] if current_index != len(experiment_paths) - 1 else None,
        prev_id=experiment_paths[current_index - 1] if current_index != 0 else None,
        hps=meta["hyperparameters"],
        env_meta=meta["environment"],
        reward_function=meta["reward_function"] if "reward_function" in meta else {},
        iterations=meta["iterations"] if "iterations" in meta else None
    )

    plots = {"normalization": {}}

    stats_path = os.path.join(path, "statistics.json")
    if os.path.isfile(stats_path):
        with open(stats_path, "r") as f:
            stats = json.load(f)

        info.update(dict(
            statistics=stats
        ))

        if "used_memory" in stats:
            plots["mem_usage"] = plot_memory_usage(stats["used_memory"])

        if "cycle_timings" in stats:
            plots["timings"] = plot_execution_times(stats["cycle_timings"],
                                                    stats.get("optimization_timings"),
                                                    stats.get("gathering_timings"))

    if "RewardNormalizationTransformer" in progress["preprocessors"]:
        plots["normalization"]["reward"] = plot_preprocessor(progress["preprocessors"]["RewardNormalizationTransformer"])

    if "StateNormalizationTransformer" in progress["preprocessors"]:
        plots["normalization"]["state"] = plot_preprocessor(progress["preprocessors"]["StateNormalizationTransformer"])

    reward_threshold = None if meta["environment"]["reward_threshold"] == "None" else float(
        meta["environment"]["reward_threshold"])

    plots["reward_distribution"] = plot_distribution(progress["rewards"]["last_cycle"], "Rewards (Last Cycle)", color=0)
    plots["length_distribution"] = plot_distribution(progress["lengths"]["last_cycle"], "Episode Lengths (Last Cycle)", color=1)
    cycles_loaded = []
    if "loaded_at" in stats.keys():
        cycles_loaded = stats["loaded_at"]
    plots["reward_progress"] = plot_reward_progress(progress["rewards"], cycles_loaded, reward_threshold=reward_threshold)
    plots["length_progress"] = plot_length_progress(progress["lengths"], cycles_loaded)
    plots["policy_loss"] = plot_loss(progress["ploss"], progress["rewards"]["mean"], "Policy Loss", color_id=0)
    plots["value_loss"] = plot_loss(progress["vloss"], progress["rewards"]["mean"], "Value Loss", color_id=1)
    plots["entropies"] = plot_loss(progress["entropies"], progress["rewards"]["mean"], "Entropy", color_id=2)

    if "per_receptor_mean" in stats.keys():
        plots["per_receptor_mean"] = plot_per_receptor_mean(stats["per_receptor_mean"])

    info.update(dict(
        plots=plots
    ))

    return flask.render_template("experiment.html", info=info)


@app.route("/compare/", methods=("POST", "GET"))
def show_comparison():
    """Compare experiments of given IDs."""
    ids = request.args.getlist('ids')

    info = {
        "ids": ids,
        "plots": {}
    }

    progress_reports = {}
    metas = {}
    for exp_id in ids:
        path = f"{PATH_TO_EXPERIMENTS}/{exp_id}"
        with open(os.path.join(path, "progress.json"), "r") as f:
            progress = json.load(f)

        with open(os.path.join(path, "meta.json"), "r") as f:
            meta = json.load(f)

        progress_reports.update({f"{exp_id}": progress})
        metas[str(exp_id)] = meta

    info["plots"]["reward"] = compare_reward_progress(
        {id: progress["rewards"] for id, progress in progress_reports.items()},
        None
    )

    info["plots"]["reward_grouped"] = grouped_reward_progress(
        {id: (metas[str(id)]["hyperparameters"]["epochs_per_cycle"], progress["rewards"]) for id, progress in progress_reports.items() if metas[str(id)]["hyperparameters"]["epochs_per_cycle"] > 30},
        None
    )

    return flask.render_template("compare.html", info=info)


@app.route("/expfile/<int:exp_id>/<path:filename>")
def expfile(exp_id, filename):
    path = os.path.abspath(os.path.join(PATH_TO_EXPERIMENTS, str(exp_id)))
    return send_from_directory(path, filename, as_attachment=True)


@app.route("/_clear_all_empty")
def clear_all_empty():
    """Delete all experiments stored that have less than 2 episodes finished."""
    experiment_paths = [os.path.join(PATH_TO_EXPERIMENTS, p) for p in os.listdir(PATH_TO_EXPERIMENTS)]

    deleted = 0
    for path in experiment_paths:
        if re.match("[0-9]+", str(path.split("/")[-1])):
            if os.path.isfile(os.path.join(path, "progress.json")):
                try:
                    with open(os.path.join(path, "progress.json"), "r") as f:
                        progress = json.load(f)
                except Exception:
                    # delete corrupted
                    shutil.rmtree(path)
                    deleted += 1
                    continue

                if "rewards" in progress and len(progress["rewards"]["mean"]) < 2:
                    shutil.rmtree(path)
                    deleted += 1

    return {"deleted": deleted}


@app.route("/_clear_all_short")
def clear_all_short():
    """Delete all experiments stored that have less than 10 cycles finished."""
    experiment_paths = [os.path.join(PATH_TO_EXPERIMENTS, p) for p in os.listdir(PATH_TO_EXPERIMENTS)]

    deleted = 0
    for path in experiment_paths:
        if re.match("[0-9]+", str(path.split("/")[-1])):
            if os.path.isfile(os.path.join(path, "progress.json")):
                try:
                    with open(os.path.join(path, "progress.json"), "r") as f:
                        progress = json.load(f)
                except Exception:
                    # delete corrupted
                    shutil.rmtree(path)
                    deleted += 1
                    continue

                if "rewards" in progress and len(progress["rewards"]["mean"]) <= 30:
                    shutil.rmtree(path)
                    deleted += 1

    return {"deleted": deleted}


@app.route('/delete_experiment', methods=['GET', 'POST'])
def delete_experiment():
    """Delete an experiment of posted id."""
    if request.method == "POST":
        try:
            eid = request.json['id']
            eid = int(eid)  # make sure its a number
            shutil.rmtree(os.path.join(PATH_TO_EXPERIMENTS, str(eid)))
        except Exception:
            return {"success": False}
    else:
        return {"success": False}

    return {"success": True}


@app.route("/evaluate", methods=("POST", "GET"))
def evaluate():
    """Evaluate an agent."""
    if request.method == "POST":
        try:
            agent = PPOAgent.from_agent_state(request.json['id'])
            evaluation_stats, _ = agent.evaluate(10, save=True)

            return {"results": evaluation_stats._asdict()}

        except Exception as e:
            return {"success": e.__repr__()}

    return {"success": "success"}



#!/usr/bin/env python
"""Pretrain the visual component."""
import argparse
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

from typing import Union

import argcomplete
import tensorflow as tf
import tensorflow_datasets as tfds

from angorapy.models import _build_openai_encoder, _build_visual_decoder, _build_openai_small_encoder
from common.const import PRETRAINED_COMPONENTS_PATH, VISION_WH
from utilities.data_generation import gen_cube_quats_prediction_data, gen_hand_pos_prediction_data
from common.loss import EuclideanDistanceLoss


def load_caltech():
    # load dataset
    test_train = tfds.load("caltech101", shuffle_files=True)
    return test_train["train"], test_train["test"], 102


def top_5_accuracy(y_true, y_pred):
    return tf.keras.metrics.top_k_categorical_accuracy(y_true, y_pred, k=5)


class TestCallback(tf.keras.callbacks.Callback):

    def __init__(self, test_data):
        self.test_data = test_data

    def on_epoch_end(self, epoch, logs={}):
        out = self.model.evaluate(self.test_data, )
        print(f"\n{out}\n")


def pretrain_on_reconstruction(pretrainable_component: Union[tf.keras.Model, str], epochs, name="visual_r"):
    """Pretrain a visual component on the reconstruction of images."""
    input_shape = pretrainable_component.input_shape
    spatial_dimensions = input_shape[1:3]

    X, _ = gen_cube_quats_prediction_data(1024 * 8)

    # model is constructed from visual component and a decoder
    decoder = _build_visual_decoder(pretrainable_component.output_shape[-1])
    model = tf.keras.Sequential((
        pretrainable_component,
        decoder
    ))

    if isinstance(pretrainable_component, tf.keras.Model):
        checkpoint_path = PRETRAINED_COMPONENTS_PATH + "/ckpts/weights.ckpt"

        # Create a callback that saves the model'serialization weights
        cp_callback = tf.keras.callbacks.ModelCheckpoint(filepath=checkpoint_path,
                                                         save_weights_only=True,
                                                         verbose=1)

        optimizer = tf.keras.optimizers.Adam()
        model.compile(optimizer, loss="mse")

        # train and save encoder
        model.fit(x=X, y=X, epochs=epochs, callbacks=[cp_callback])
        pretrainable_component.save(PRETRAINED_COMPONENTS_PATH + f"/{name}.h5")
        model.save(PRETRAINED_COMPONENTS_PATH + f"/{name}_full.h5")
    elif isinstance(pretrainable_component, str):
        model.load_weights(pretrainable_component)
    else:
        raise ValueError("No clue what you think this is but it for sure ain't no model nor a path to model.")


def pretrain_on_classification(pretrainable_component: Union[tf.keras.Model, str], epochs, name="visual_c"):
    """Pretrain a visual component on the classification of images."""

    input_shape = pretrainable_component.input_shape
    spatial_dimensions = input_shape[1:3]

    train_images, test_images, n_classes = load_caltech()

    # resize and normalize images, extract one hot vectors from labels
    train_images = train_images.map(
        lambda img: (tf.image.resize(img["image"], spatial_dimensions) / 255,
                     tf.one_hot(img["label"], depth=n_classes)))
    train_images = train_images.batch(128)

    test_images = test_images.map(
        lambda img: (tf.image.resize(img["image"], spatial_dimensions) / 255,
                     tf.one_hot(img["label"], depth=n_classes)))
    test_images = test_images.batch(128)

    # model is constructed from visual component and classification layer
    model = tf.keras.Sequential((
        pretrainable_component,
        tf.keras.layers.Dense(n_classes),
        tf.keras.layers.Activation("softmax")
    ))

    if isinstance(pretrainable_component, tf.keras.Model):
        optimizer = tf.keras.optimizers.Adam(learning_rate=0.001)
        model.compile(optimizer, loss="categorical_crossentropy", metrics=["accuracy"])

        # train and save encoder
        model.fit(train_images, epochs=epochs, callbacks=[])
        pretrainable_component.save(PRETRAINED_COMPONENTS_PATH + f"/{name}.h5")
        model.save(PRETRAINED_COMPONENTS_PATH + f"/{name}_full.h5")
    elif isinstance(pretrainable_component, str):
        model.load_weights(pretrainable_component)
    else:
        raise ValueError("No clue what you think this is but it for sure ain't no model nor a path to model.")

    # evaluate
    results = model.evaluate(test_images)
    print(results)


def pretrain_on_object_pose(pretrainable_component: Union[tf.keras.Model, str], epochs, name="visual_op"):
    """Pretrain a visual component on prediction of cube position."""

    # generate training data
    X, Y = gen_cube_quats_prediction_data(1024 * 8)
    X = tf.image.per_image_standardization(X)

    # model is constructed from visual component and regression output
    model = tf.keras.Sequential((
        pretrainable_component,
        tf.keras.layers.Dense(7, activation="linear"),
    ))

    if isinstance(pretrainable_component, tf.keras.Model):
        optimizer = tf.keras.optimizers.Adam(learning_rate=0.001)
        model.compile(optimizer, loss="mse", metrics=[EuclideanDistanceLoss])

        # train and save encoder
        model.fit(X, Y, epochs=epochs, callbacks=[], batch_size=128)
        pretrainable_component.save(PRETRAINED_COMPONENTS_PATH + f"/{name}.h5")
    elif isinstance(pretrainable_component, str):
        model.load_weights(pretrainable_component)
    else:
        raise ValueError("No clue what you think this is but it for sure ain't no model nor a path to model.")


def pretrain_on_hand_pose(pretrainable_component: Union[tf.keras.Model, str], epochs, name="visual_hp"):
    """Pretrain a visual component on prediction of cube position."""

    # model is constructed from visual component and regression output
    model = pretrainable_component

    if isinstance(pretrainable_component, tf.keras.Model):
        optimizer = tf.keras.optimizers.Adam(learning_rate=0.0005)

        for epoch_i in range(epochs):
            print(f"Starting Epoch {epoch_i}")

            # generate training data
            X, Y = gen_hand_pos_prediction_data(1024 * 8)
            X = tf.image.per_image_standardization(X)

            dataset = tf.data.Dataset.from_tensor_slices((X, Y))
            dataset = dataset.shuffle(buffer_size=1024).batch(128)

            for step, (batch_x, batch_y) in enumerate(dataset):
                with tf.GradientTape() as tape:
                    output = model(batch_x, training=True)
                    loss = tf.keras.losses.MSE(batch_y, output)

                grads = tape.gradient(loss, model.trainable_weights)
                optimizer.apply_gradients(zip(grads, model.trainable_weights))

                if step % 4 == 0:
                    print(f"Training loss (for one batch) at step {step}: {float(tf.reduce_mean(loss))}. "
                          f"Seen so far: {((step + 1) * 128)} samples")

            del dataset

        # save encoder
        pretrainable_component.save(PRETRAINED_COMPONENTS_PATH + f"/{name}.h5")
    elif isinstance(pretrainable_component, str):
        model.load_weights(pretrainable_component)
    else:
        raise ValueError("No clue what you think this is but it for sure ain't no model nor a path to a model.")


if __name__ == "__main__":
    tf.get_logger().setLevel('INFO')

    # parse commandline arguments
    parser = argparse.ArgumentParser(description="Pretrain a visual component on classification or reconstruction.")

    # general parameters
    parser.add_argument("task", nargs="?", type=str, choices=["classify", "reconstruct", "hand", "object",
                                                              "c", "r", "h", "o", "hp", "op"], default="h")
    parser.add_argument("--name", type=str, default="pretrained_component",
                        help="Name the pretraining to uniquely identify it.")
    parser.add_argument("--load", type=str, default=None, help=f"load the weights from checkpoint path")
    parser.add_argument("--epochs", type=int, default=60, help=f"number of pretraining epochs")

    # read arguments
    argcomplete.autocomplete(parser)
    args = parser.parse_args()

    visual_component = _build_openai_small_encoder(shape=(VISION_WH, VISION_WH, 3), out_shape=15, name="visual_component")
    visual_component(tf.random.normal((16, VISION_WH, VISION_WH, 3)))  # needed to initialize keras or whatever

    os.makedirs(PRETRAINED_COMPONENTS_PATH, exist_ok=True)

    args.name = args.name + "_" + args.task[0]

    if args.task in ["classify", "c"]:
        pretrain_on_classification(visual_component, args.epochs, name=args.name)
    elif args.task in ["reconstruct", "r"]:
        pretrain_on_reconstruction(visual_component, args.epochs, name=args.name)
    elif args.task in ["hand", "h"]:
        pretrain_on_hand_pose(visual_component, args.epochs, name=args.name)
    else:
        raise ValueError("I dont know that task type.")



"""A collection of builder functions for models usable with different environments in PPO."""
from typing import Callable

from angorapy.models.components import _build_fcn_component, _build_encoding_sub_model
from angorapy.models.convolutional import _build_visual_decoder, _build_openai_encoder
from angorapy.models.mighty_maker import *
from angorapy.models.shadow import build_shadow_brain_models, build_shadow_brain_base
from angorapy.models.simple import build_ffn_models, build_deeper_models, build_rnn_models, _build_encoding_sub_model, \
    build_simple_models


def get_model_builder(model="simple", model_type: str = "ffn", shared: bool = False, blind: bool = True) -> Callable:
    """Get a builder function for a model with the described parameters."""
    # TODO shared seems not to work yet
    params = locals()
    params["shared"] = "shared" if params["shared"] else "distinct"
    params["blind"] = "blind" if params["blind"] else ""

    params = [str(val) for key, val in sorted(params.items(), key=lambda x: x[0]) if val != ""]

    return globals()[f"build_{'_'.join(params)}_models"]



import os
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

from angorapy.analysis.investigation import Investigator

try:
    from mpi4py import MPI

    is_root = MPI.COMM_WORLD.rank == 0
except:
    is_root = True

from angorapy.agent.ppo_agent import PPOAgent
from angorapy.common.policies import BetaPolicyDistribution, MultiCategoricalPolicyDistribution
from angorapy.common.transformers import RewardNormalizationTransformer, StateNormalizationTransformer
from angorapy.common.wrappers import make_env

from angorapy.models import get_model_builder

import unittest


def _test_drill(env_name, model_builder=None,):
    """Test drilling an agent."""
    wrappers = [StateNormalizationTransformer, RewardNormalizationTransformer]
    env = make_env(env_name, reward_config=None, transformers=wrappers)
    if model_builder is None:
        build_models = get_model_builder(model="simple", model_type="ffn", shared=False)
    else:
        build_models = model_builder
    agent = PPOAgent(build_models, env, workers=2, horizon=128)
    agent.drill(n=2, epochs=2, batch_size=64)


class AgentTest(unittest.TestCase):

    def test_drill_continuous(self):
        """Test drilling of continuous agent (LunarLanderContinuous)."""

        try:
            _test_drill("LunarLanderContinuous-v2")
        except Exception:
            self.fail("Continuous drill raises error.")

    def test_drill_discrete(self):
        """Test drilling of discrete agent (LunarLander)."""

        try:
            _test_drill("LunarLander-v2")
        except Exception:
            self.fail("Discrete drill raises error.")

    def test_drill_manipulate_multicontinuous(self):
        """Test drilling of discrete agent (LunarLander)."""

        try:
            wrappers = [StateNormalizationTransformer, RewardNormalizationTransformer]
            env = make_env("HumanoidManipulateBlockDiscreteAsynchronous-v0", reward_config=None, transformers=wrappers)
            build_models = get_model_builder(model="shadow", model_type="lstm", shared=False)
            agent = PPOAgent(build_models, env, workers=2, horizon=128, distribution=MultiCategoricalPolicyDistribution(env))
            agent.drill(n=2, epochs=2, batch_size=64)
        except Exception:
            self.fail("HumanoidManipulateBlockDiscreteAsynchronous drill raises error.")

    def test_drill_manipulate_continuous(self):
        """Test drilling of discrete agent (LunarLander)."""

        try:
            wrappers = [StateNormalizationTransformer, RewardNormalizationTransformer]
            env = make_env("HumanoidManipulateBlockAsynchronous-v0", reward_config=None, transformers=wrappers)
            build_models = get_model_builder(model="shadow", model_type="lstm", shared=False)
            agent = PPOAgent(build_models, env, workers=2, horizon=128, distribution=BetaPolicyDistribution(env))
            agent.drill(n=2, epochs=2, batch_size=64)
        except Exception:
            self.fail("HumanoidManipulateBlockDiscreteAsynchronous drill raises error.")

    def test_drill_reach(self):
        """Test drilling of discrete agent (LunarLander)."""

        try:
            wrappers = [StateNormalizationTransformer, RewardNormalizationTransformer]
            env = make_env("ReachAbsolute-v0", reward_config=None, transformers=wrappers)
            build_models = get_model_builder(model="shadow", model_type="lstm", shared=False)
            agent = PPOAgent(build_models, env, workers=2, horizon=128, distribution=BetaPolicyDistribution(env))
            agent.drill(n=2, epochs=2, batch_size=64)
        except Exception:
            self.fail("ReachAbsolute drill raises error.")

    def test_drill_freereach(self):
        """Test drilling of free reach agent."""

        try:
            wrappers = [StateNormalizationTransformer, RewardNormalizationTransformer]
            env = make_env("FreeReachAbsolute-v0", reward_config=None, transformers=wrappers)
            build_models = get_model_builder(model="shadow", model_type="lstm", shared=False)
            agent = PPOAgent(build_models, env, workers=2, horizon=128, distribution=BetaPolicyDistribution(env))
            agent.drill(n=2, epochs=2, batch_size=64)
        except Exception:
            self.fail("FreeReachAbsolute drill raises error.")


if __name__ == '__main__':
    unittest.main()



import json
import os

from matplotlib import pyplot as plt

from common.const import PATH_TO_EXPERIMENTS

experiment_ids = [   1626374994   ]

reward_developments = {}
for id in experiment_ids:
    with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "meta.json")) as f:
        meta = json.load(f)
    with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "progress.json")) as f:
        progress = json.load(f)

    exp_name = meta["hyperparameters"]["model"].upper()
    reward_developments[exp_name] = progress["rewards"]["mean"]

for name, rewards in reward_developments.items():
    plt.plot(rewards, label=name)

plt.xlabel("Cycle")
plt.ylabel("Reward")
# plt.legend()

plt.gcf().set_size_inches(8, 4)
# plt.show()
plt.savefig("manipulate-progress.pdf", format="pdf", bbox_inches="tight")


#!/usr/bin/env python
"""Evaluate a loaded agent on a task."""

import argparse
import os

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

from environments.shadowhand import BaseShadowHandEnv

from utilities.datatypes import mpi_condense_stats
import statistics
import time

from agent.ppo_agent import PPOAgent
from common.const import BASE_SAVE_PATH
from mpi4py import MPI
import tensorflow as tf

tf.get_logger().setLevel('INFO')

parser = argparse.ArgumentParser(description="Evaluate an agent.")
parser.add_argument("id", type=int, nargs="?", help="id of the agent, defaults to newest", default=None)
parser.add_argument("-n", type=int, help="number of evaluation episodes", default=10)
parser.add_argument("--act-confidently", action="store_true", help="act deterministically without stochasticity")
args = parser.parse_args()

mpi_comm = MPI.COMM_WORLD
is_root = mpi_comm.rank == 0

if args.id is None:
    ids = map(int, os.listdir(BASE_SAVE_PATH))
    args.id = max(ids)

start = time.time()
agent = PPOAgent.from_agent_state(args.id, "b")

if isinstance(agent.env.unwrapped, BaseShadowHandEnv):
    agent.env.env.set_delta_t_simulation(0.0005)

if is_root:
    print(f"Agent {args.id} successfully loaded from state 'best' (training performance: {agent.cycle_reward_history[-1]}).")


# determine number of repetitions on this worker
worker_base, worker_extra = divmod(args.n, MPI.COMM_WORLD.size)
worker_split = [worker_base + (r < worker_extra) for r in range(MPI.COMM_WORLD.size)]
workers_n = worker_split[mpi_comm.rank]

stat_bundles = []
for _ in range(workers_n):
    stats, _ = agent.evaluate(1, act_confidently=args.act_confidently)
    stat_bundles.append(stats)

stats = mpi_condense_stats(stat_bundles)

if is_root:
    average_reward = round(statistics.mean(stats.episode_rewards), 2)
    average_length = round(statistics.mean(stats.episode_lengths), 2)
    std_reward = round(statistics.stdev(stats.episode_rewards), 2)
    std_length = round(statistics.stdev(stats.episode_lengths), 2)

    print(f"Evaluated agent on {stats.numb_completed_episodes} x {agent.env_name} and achieved an average reward of {average_reward} [std: {std_reward}; "
          f"between ({min(stats.episode_rewards)}, {max(stats.episode_rewards)})].\n "
          f"The agent {'acted confidently' if args.act_confidently else 'acted stochastically.'}\n"
          f"An episode on average took {average_length} steps [std: {std_length}; "
          f"between ({min(stats.episode_lengths)}, {max(stats.episode_lengths)})].\n"
          f"This took me {round(time.time() - start, 2)}s.")



import os

from common.wrappers import make_env

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import random
import unittest

import gym
import numpy as np
import tensorflow as tf
from scipy.stats import norm, entropy, beta

from agent.core import extract_discrete_action_probabilities
from common.policies import GaussianPolicyDistribution, CategoricalPolicyDistribution, BetaPolicyDistribution
from common.const import NP_FLOAT_PREC
from common.senses import Sensation
from common.transformers import RewardNormalizationTransformer, StateNormalizationTransformer, merge_transformers
from utilities.model_utils import reset_states_masked
from utilities.util import env_extract_dims



class CoreTest(unittest.TestCase):

    def test_extract_discrete_action_probabilities(self):
        # no recurrence
        action_probs = tf.convert_to_tensor([[1, 5], [3, 7], [7, 2], [8, 4], [0, 2], [4, 5], [4, 2], [7, 5]])
        actions = tf.convert_to_tensor([1, 0, 1, 1, 0, 0, 0, 1])
        result_reference = tf.convert_to_tensor([5, 3, 2, 4, 0, 4, 4, 5])
        result = extract_discrete_action_probabilities(action_probs, actions)

        self.assertTrue(tf.reduce_all(tf.equal(result, result_reference)).numpy().item())

    def test_extract_discrete_action_probabilities_with_recurrence(self):
        tf.config.experimental_run_functions_eagerly(True)

        # with recurrence
        action_probs = tf.convert_to_tensor(
            [[[1, 5], [1, 5]], [[3, 7], [3, 7]], [[7, 2], [7, 2]], [[8, 4], [8, 4]], [[0, 2], [0, 2]], [[4, 5], [4, 5]],
             [[4, 2], [4, 2]], [[7, 5], [7, 5]]])
        actions = tf.convert_to_tensor([[1, 1], [0, 0], [1, 1], [1, 1], [0, 0], [0, 0], [0, 0], [1, 1]])
        result_reference = tf.convert_to_tensor([[5, 5], [3, 3], [2, 2], [4, 4], [0, 0], [4, 4], [4, 4], [5, 5]])
        result = extract_discrete_action_probabilities(action_probs, actions)

        self.assertTrue(tf.reduce_all(tf.equal(result, result_reference)).numpy().item())


class ProbabilityTest(unittest.TestCase):

    # GAUSSIAN

    def test_gaussian_pdf(self):
        distro = GaussianPolicyDistribution(gym.make("LunarLanderContinuous-v2"))

        x = tf.convert_to_tensor([[2, 3], [4, 3], [2, 1]], dtype=tf.float32)
        mu = tf.convert_to_tensor([[2, 1], [1, 3], [2, 2]], dtype=tf.float32)
        sig = tf.convert_to_tensor([[2, 2], [1, 2], [2, 1]], dtype=tf.float32)

        result_reference = np.prod(norm.pdf(x, loc=mu, scale=sig), axis=-1)
        result_pdf = distro.probability(x, mu, sig).numpy()
        result_log_pdf = np.exp(distro.log_probability(x, mu, np.log(sig)).numpy())

        self.assertTrue(np.allclose(result_reference, result_pdf), msg="Gaussian PDF returns wrong Result")
        self.assertTrue(np.allclose(result_pdf, result_log_pdf), msg="Gaussian Log PDF returns wrong Result")

    def test_gaussian_entropy(self):
        distro = GaussianPolicyDistribution(gym.make("LunarLanderContinuous-v2"))

        mu = tf.convert_to_tensor([[2.0, 3.0], [2.0, 1.0]], dtype=tf.float32)
        sig = tf.convert_to_tensor([[1.0, 1.0], [1.0, 5.0]], dtype=tf.float32)

        result_reference = np.sum(norm.entropy(loc=mu, scale=sig), axis=-1)
        result_log = distro.entropy(np.log(sig)).numpy()
        result = distro._entropy_from_params(sig).numpy()

        self.assertTrue(np.allclose(result_reference, result), msg="Gaussian entropy returns wrong result")
        self.assertTrue(np.allclose(result_log, result_reference), msg="Gaussian entropy from log returns wrong result")

    # BETA

    def test_beta_pdf(self):
        distro = BetaPolicyDistribution(gym.make("LunarLanderContinuous-v2"))

        x = tf.convert_to_tensor([[0.2, 0.3], [0.4, 0.3], [0.2, 0.1]], dtype=tf.float32)
        alphas = tf.convert_to_tensor([[2, 1], [1, 3], [2, 2]], dtype=tf.float32)
        betas = tf.convert_to_tensor([[2, 2], [1, 2], [2, 1]], dtype=tf.float32)

        result_reference = np.prod(beta.pdf(distro._scale_sample_to_distribution_range(x), alphas, betas), axis=-1)
        result_pdf = distro.probability(x, alphas, betas).numpy()
        result_log_pdf = np.exp(distro.log_probability(x, alphas, betas).numpy())

        self.assertTrue(np.allclose(result_reference, result_log_pdf), msg="Beta Log PDF returns wrong Result")
        self.assertTrue(np.allclose(result_reference, result_pdf), msg="Beta PDF returns wrong Result")

    def test_beta_entropy(self):
        distro = BetaPolicyDistribution(gym.make("LunarLanderContinuous-v2"))

        alphas = tf.convert_to_tensor([[2, 1], [1, 3], [2, 2]], dtype=tf.float32)
        betas = tf.convert_to_tensor([[2, 2], [1, 2], [2, 1]], dtype=tf.float32)

        result_reference = np.sum(beta.entropy(alphas, betas), axis=-1)
        result_pdf = distro._entropy_from_params((alphas, betas)).numpy()

        self.assertTrue(np.allclose(result_reference, result_pdf), msg="Beta PDF returns wrong Result")

    # CATEGORICAL

    def test_categorical_entropy(self):
        distro = CategoricalPolicyDistribution(gym.make("CartPole-v1"))

        probs = tf.convert_to_tensor([[0.1, 0.4, 0.2, 0.25, 0.05],
                                      [0.1, 0.4, 0.2, 0.2, 0.1],
                                      [0.1, 0.35, 0.3, 0.24, 0.01]], dtype=tf.float32)

        result_reference = [entropy(probs[i]) for i in range(len(probs))]
        result_log = distro._entropy_from_log_pmf(np.log(probs)).numpy()
        result = distro._entropy_from_pmf(probs).numpy()

        self.assertTrue(np.allclose(result_reference, result), msg="Discrete entropy returns wrong result")
        self.assertTrue(np.allclose(result_log, result_reference), msg="Discrete entropy from log returns wrong result")


class UtilTest(unittest.TestCase):

    def test_masked_state_reset(self):
        model = tf.keras.Sequential((
            tf.keras.layers.Dense(2, batch_input_shape=(7, None, 2)),
            tf.keras.layers.LSTM(5, stateful=True, name="larry", return_sequences=True),
            tf.keras.layers.LSTM(5, stateful=True, name="harry"))
        )

        l_layer = model.get_layer("larry")
        h_layer = model.get_layer("harry")
        l_layer.reset_states([s.numpy() + 9 for s in l_layer.states])
        h_layer.reset_states([s.numpy() + 9 for s in h_layer.states])
        reset_states_masked(model, [True, False, False, True, False, False, True])

        self.assertTrue(np.allclose([s.numpy() for s in model.get_layer("larry").states],
                                    [s.numpy() for s in model.get_layer("harry").states]))
        self.assertTrue(np.allclose([s.numpy() for s in model.get_layer("larry").states], [
            [0, 0, 0, 0, 0],
            [9, 9, 9, 9, 9],
            [9, 9, 9, 9, 9],
            [0, 0, 0, 0, 0],
            [9, 9, 9, 9, 9],
            [9, 9, 9, 9, 9],
            [0, 0, 0, 0, 0],
        ]))


class WrapperTest(unittest.TestCase):

    def test_state_normalization(self):
        env_name = "LunarLanderContinuous-v2"
        env = make_env(env_name)
        normalizer = StateNormalizationTransformer(env_name, *env_extract_dims(env))

        inputs = [env.observation_space.sample() for _ in range(15)]
        true_mean = np.mean(inputs, axis=0)
        true_std = np.std(inputs, axis=0)

        for sample in inputs:
            o, _, _, _ = normalizer.transform((Sensation(proprioception=sample), 1.0, False, {}))

        self.assertTrue(np.allclose(true_mean, normalizer.mean["proprioception"]))
        self.assertTrue(np.allclose(true_std, np.sqrt(normalizer.variance["proprioception"])))

    def test_reward_normalization(self):
        env_name = "LunarLanderContinuous-v2"
        env = make_env(env_name)
        normalizer = RewardNormalizationTransformer(env_name, *env_extract_dims(env))

        inputs = [random.random() * 10 for _ in range(1000)]
        true_mean = np.mean(inputs, axis=0)
        true_std = np.std(inputs, axis=0)

        for sample in inputs:
            o, _, _, _ = normalizer.transform((1, sample, 1, 1))

        self.assertTrue(np.allclose(true_mean, normalizer.mean["reward"]))
        self.assertTrue(np.allclose(true_std, np.sqrt(normalizer.variance["reward"])))

    def test_state_normalization_adding(self):
        env_name = "LunarLanderContinuous-v2"
        env = make_env(env_name)
        normalizer_a = StateNormalizationTransformer(env_name, *env_extract_dims(env))
        normalizer_b = StateNormalizationTransformer(env_name, *env_extract_dims(env))
        normalizer_c = StateNormalizationTransformer(env_name, *env_extract_dims(env))

        inputs_a = [env.observation_space.sample() for _ in range(100)]
        inputs_b = [env.observation_space.sample() for _ in range(100)]
        inputs_c = [env.observation_space.sample() for _ in range(100)]

        true_mean = np.mean(inputs_a + inputs_b + inputs_c, axis=0)
        true_std = np.std(inputs_a + inputs_b + inputs_c, axis=0)

        for sample in inputs_a:
            normalizer_a.update({"proprioception": sample})

        for sample in inputs_b:
            normalizer_b.update({"proprioception": sample})

        for sample in inputs_c:
            normalizer_c.update({"proprioception": sample})

        combined_normalizer = normalizer_a + normalizer_b + normalizer_c
        merged_normalizer = merge_transformers([normalizer_a, normalizer_b, normalizer_c])

        # print(normalizer_a.mean["proprioception"])
        # print(normalizer_b.mean["proprioception"])
        # print(normalizer_c.mean["proprioception"])
        # print(true_mean)
        # print(combined_normalizer.mean["proprioception"])

        self.assertTrue(np.allclose(true_mean, combined_normalizer.mean["proprioception"]))
        self.assertTrue(np.allclose(true_mean, merged_normalizer.mean["proprioception"]))
        self.assertTrue(np.allclose(true_std, np.sqrt(combined_normalizer.variance["proprioception"])))

    def test_reward_normalization_adding(self):
        env_name = "LunarLanderContinuous-v2"
        env = make_env(env_name)
        normalizer_a = RewardNormalizationTransformer(env_name, *env_extract_dims(env))
        normalizer_b = RewardNormalizationTransformer(env_name, *env_extract_dims(env))
        normalizer_c = RewardNormalizationTransformer(env_name, *env_extract_dims(env))

        inputs_a = [random.random() * 10 for _ in range(1000)]
        inputs_b = [random.random() * 20 for _ in range(1000)]
        inputs_c = [random.random() * 5 for _ in range(1000)]

        true_mean = np.mean(inputs_a + inputs_b + inputs_c, axis=0)
        true_std = np.std(inputs_a + inputs_b + inputs_c, axis=0)

        for sample in inputs_a:
            normalizer_a.update({"reward": sample})

        for sample in inputs_b:
            normalizer_b.update({"reward": sample})

        for sample in inputs_c:
            normalizer_c.update({"reward": sample})

        combined_normalizer = normalizer_a + normalizer_b + normalizer_c

        self.assertTrue(np.allclose(true_mean, combined_normalizer.mean["reward"]))
        self.assertTrue(np.allclose(true_std, np.sqrt(combined_normalizer.variance["reward"])))


if __name__ == '__main__':
    tf.config.experimental_run_functions_eagerly(True)

    testsuite = unittest.TestLoader().discover('.')
    unittest.TextTestRunner(verbosity=1).run(testsuite)


import unittest

import gym
import numpy as np
import tensorflow as tf

from common.policies import BetaPolicyDistribution
from angorapy.models import get_model_builder
from utilities.model_utils import reset_states_masked, build_sub_model_from, get_layers_by_names


class UtilTest(unittest.TestCase):

    def test_masked_state_reset(self):
        model = tf.keras.Sequential((
            tf.keras.layers.Dense(2, batch_input_shape=(7, None, 2)),
            tf.keras.layers.LSTM(5, stateful=True, name="larry", return_sequences=True),
            tf.keras.layers.LSTM(5, stateful=True, name="harry"))
        )

        l_layer = model.get_layer("larry")
        h_layer = model.get_layer("harry")
        l_layer.reset_states([s.numpy() + 9 for s in l_layer.states])
        h_layer.reset_states([s.numpy() + 9 for s in h_layer.states])
        reset_states_masked(model, [True, False, False, True, False, False, True])

        self.assertTrue(np.allclose([s.numpy() for s in model.get_layer("larry").states],
                                    [s.numpy() for s in model.get_layer("harry").states]))
        self.assertTrue(np.allclose([s.numpy() for s in model.get_layer("larry").states], [
            [0, 0, 0, 0, 0],
            [9, 9, 9, 9, 9],
            [9, 9, 9, 9, 9],
            [0, 0, 0, 0, 0],
            [9, 9, 9, 9, 9],
            [9, 9, 9, 9, 9],
            [0, 0, 0, 0, 0],
        ]))

    def test_submodeling_from(self):
        env = gym.make("LunarLanderContinuous-v2")
        full_model, _, _ = get_model_builder("simple", "gru", shared=False)(env, BetaPolicyDistribution(env))
        sub_model_from_a = build_sub_model_from(full_model, "beta_action_head")
        sub_model_from_b = build_sub_model_from(full_model, "policy_recurrent_layer")

        for sub_model_from in [sub_model_from_a, sub_model_from_b]:
            layer = get_layers_by_names(sub_model_from, ["beta_action_head"])[0]

            input_shape_raw = layer.get_input_shape_at(1)
            input_shape_replaced = tuple(v if v is not None else 1 for v in input_shape_raw)

            out = sub_model_from(tf.random.normal(input_shape_replaced))


import tensorflow as tf

from angorapy.common.wrappers import BaseWrapper, make_env
from angorapy.utilities.error import IncompatibleModelException


def validate_env_model_compatibility(env: BaseWrapper, model: tf.keras.Model) -> bool:
    """Validate whether the environments states can be processed by the model."""
    model_inputs = [ip.name for ip in model.inputs]
    env_outputs = list(env.reset().dict().keys())

    if not sorted(model_inputs) == sorted(env_outputs):
        raise IncompatibleModelException(
            f"The model with inputs {model_inputs} cannot handle this environment's states with senses {env_outputs}."
        )

    return True


if __name__ == '__main__':
    from environments import *
    from common.policies import BetaPolicyDistribution
    from angorapy.models import build_shadow_brain_base, build_simple_models

    e = make_env("LunarLanderContinuous-v2")
    _, _, m = build_simple_models(e, BetaPolicyDistribution(e), bs=16)
    print(validate_env_model_compatibility(e, m))

    e = make_env("ReachAbsolute-v0")
    _, _, m = build_shadow_brain_base(e, BetaPolicyDistribution(e), bs=16)
    print(validate_env_model_compatibility(e, m))




from typing import List, Dict, Tuple

import bokeh
import numpy
import numpy as np
import pandas as pd
import scipy
from bokeh import embed
from bokeh.models import Span, Range1d, LinearAxis, ColumnDataSource, HoverTool
from bokeh.plotting import figure
from scipy import stats

from angorapy.utilities.monitor.plotting_base import palette, plot_styling, style_plot
from angorapy.utilities.statistics import mean_fill_nones


def plot_memory_usage(memory_trace: List):
    """Plot the development of memory used by the training."""
    x = list(range(len(memory_trace)))
    y = memory_trace

    p = figure(title="Memory Usage",
               x_axis_label='Cycle',
               y_axis_label='Memory in GB',
               y_range=(min(memory_trace), max(memory_trace)),
               **plot_styling)

    p.line(x, y, legend_label="RAM", line_width=2, color=palette[0])

    p.legend.location = "bottom_right"
    style_plot(p)

    return embed.components(p)


def plot_execution_times(cycle_timings, optimization_timings=None, gathering_timings=None):
    """Plot the execution times of a full cycle and optionally bot sub phases."""
    x = list(range(len(cycle_timings)))

    all_times = (cycle_timings
                 + (optimization_timings if optimization_timings is not None else [])
                 + (gathering_timings if gathering_timings is not None else []))

    if len(all_times) < 1:
        return None, None

    p = figure(title="Execution Times",
               x_axis_label='Cycle',
               y_axis_label='Seconds',
               y_range=(0, max(all_times)),
               **plot_styling)

    p.line(x, cycle_timings, legend_label="Full Cycle", line_width=2, color=palette[0])
    if optimization_timings is not None:
        p.line(x, optimization_timings, legend_label="Optimization Phase", line_width=2, color=palette[1])
    if gathering_timings is not None:
        p.line(x, gathering_timings, legend_label="Gathering Phase", line_width=2, color=palette[2])

    p.legend.location = "bottom_right"
    style_plot(p)

    return embed.components(p)


def plot_reward_progress(rewards: Dict[str, list], cycles_loaded, reward_threshold=None):
    """Plot the execution times of a full cycle and optionally bot sub phases."""
    means, stds = rewards["mean"], rewards["stdev"]
    stds = np.array(stds)
    stds[stds == None] = 0
    stds = stds * 0.2
    means = np.array(means)
    means = mean_fill_nones(means)

    x = list(range(len(means)))
    df = pd.DataFrame(data=dict(x=x, y=means, lower=np.subtract(means, stds), upper=np.add(means, stds)))
    value_range = max(df["upper"]) - min(df["lower"])

    tooltip = HoverTool(
        tooltips=[("Cycle", "@x"),
                  ("Reward", "@y")],
        mode="vline"
    )

    range_max = max(df["upper"]) + value_range * 0.2
    if reward_threshold is not None:
        range_max = max(range_max, reward_threshold * 1.1)
    p = figure(title="Average Rewards per Cycle",
               x_axis_label='Cycle',
               y_axis_label='Total Episode Return',
               y_range=(min(df["lower"]), range_max),
               x_range=(0, max(x)),
               **plot_styling)

    p.add_tools(tooltip)
    p.add_tools(bokeh.models.BoxZoomTool())
    p.add_tools(bokeh.models.ResetTool())
    p.add_tools(bokeh.models.SaveTool())

    # ERROR BAND
    error_band = bokeh.models.Band(
        base="x", lower="lower", upper="upper",
        source=ColumnDataSource(df.reset_index()),
        fill_color=palette[0],
        fill_alpha=0.2,
        line_color=palette[0],
        line_alpha=0.4,
    )
    p.add_layout(error_band)
    p.renderers.extend([error_band])

    # REWARD LINE
    p.line(x, means, legend_label="Reward", line_width=2, color=palette[0])

    # REWARD THRESHOLD
    if reward_threshold is not None:
        p.line(x, [reward_threshold for _ in x], line_color="green", line_width=2, line_alpha=0.7, line_dash="dashed",
               legend_label="Solution Threshold")

    # MAX VALUE MARKING
    x_max = np.argmax(means)
    y_max = np.max(means)
    p.add_layout(bokeh.models.Arrow(end=bokeh.models.NormalHead(size=10,
                                                                line_color="darkred",
                                                                line_width=2,
                                                                fill_color="red"),
                                    line_color="darkred",
                                    line_width=2,
                                    x_start=x_max, y_start=y_max + value_range * 0.07,
                                    x_end=x_max, y_end=y_max))
    p.add_layout(bokeh.models.Label(x=x_max, y=y_max + value_range * 0.075, text=str(y_max),
                                    border_line_color='black', border_line_alpha=0,
                                    background_fill_color='white', background_fill_alpha=1.0, text_align="center",
                                    text_line_height=1.5, text_font_size="10pt", text_color="darkred"

                                    ))

    load_markings = []
    for load_timing in cycles_loaded:
        load_markings.append(
            Span(location=load_timing[0], dimension="height", line_color="red", line_width=2, line_dash=[6, 3])
        )

    p.renderers.extend(load_markings)

    p.legend.location = "bottom_right"
    style_plot(p)

    return embed.components(p)


def compare_reward_progress(rewards: Dict[str, Dict[str, list]], reward_threshold=None):
    """Plot the execution times of a full cycle and optionally bot sub phases."""
    y_min = 0
    y_max = 0
    x_all = []

    p = figure(title="Average Rewards per Cycle",
               x_axis_label='Cycle',
               y_axis_label='Total Episode Return',
               **plot_styling)
    p.x_range.start = 0

    names_and_lines = []
    for i, (name, data) in enumerate(rewards.items()):
        means, stds = data["mean"], data["stdev"]
        stds = np.array(stds)
        stds[stds == None] = 0
        stds = stds * 0.2
        means = np.array(means)
        means = mean_fill_nones(means)

        y_min = min(min(means), y_min)
        y_max = max(max(means), y_max)

        x = list(range(len(means)))
        if len(x) > len(x_all):
            x_all = x

        df = pd.DataFrame(data=dict(x=x, y=means, lower=np.subtract(means, stds), upper=np.add(means, stds)))
        value_range = max(df["upper"]) - min(df["lower"])

        range_max = max(df["upper"]) + value_range * 0.1
        if reward_threshold is not None:
            range_max = max(range_max, reward_threshold * 1.1)

        # ERROR BAND
        error_band = bokeh.models.Band(
            base="x", lower="lower", upper="upper",
            source=ColumnDataSource(df.reset_index()),
            fill_color=palette[i % len(palette)],
            fill_alpha=0.2,
            line_color=palette[i % len(palette)],
            line_alpha=0.4,
        )
        p.add_layout(error_band)
        p.renderers.extend([error_band])

        # REWARD LINE
        reward_line = p.line(x, means, line_width=2, color=palette[i % len(palette)])
        names_and_lines.append((
            name, [reward_line]
        ))

    p.add_tools(bokeh.models.BoxZoomTool())
    p.add_tools(bokeh.models.ResetTool())
    p.add_tools(bokeh.models.SaveTool())

    # REWARD THRESHOLD
    if reward_threshold is not None:
        p.line(x_all, [reward_threshold for _ in x_all], line_color="green", line_width=2, line_alpha=0.7, line_dash="dashed",
               legend_label="Solution Threshold")

    p.y_range = bokeh.models.Range1d(y_min, y_max)

    legend = bokeh.models.Legend(items=names_and_lines, location="bottom_center")
    p.add_layout(legend, "below")

    style_plot(p)
    p.height = 750 + legend.glyph_height * len(rewards)

    return embed.components(p)


def grouped_reward_progress(rewards: Dict[str, Tuple[int, Dict[str, list]]], reward_threshold=None):
    y_min = 0
    y_max = 0
    x_all = []

    p = figure(title="Average Rewards per Cycle",
               x_axis_label='Cycle',
               y_axis_label='Total Episode Return',
               **plot_styling)
    p.x_range.start = 0

    by_groups = {}
    for name, (group, data) in rewards.items():
        if group not in by_groups.keys():
            by_groups[group] = []

        by_groups[group].append(data)

    i = 0
    names_and_lines = []
    tooltip_parts = []
    for group, datas in by_groups.items():
        means = np.mean([mean_fill_nones(d["mean"]) for d in datas], axis=0)
        stds = np.std([d["mean"] for d in datas], axis=0)

        y_min = min(min(means), y_min)
        y_max = max(max(means), y_max)

        x = list(range(len(means)))
        if len(x) > len(x_all):
            x_all = x

        df = pd.DataFrame(data=dict(x=x, y=means, lower=np.subtract(means, stds), upper=np.add(means, stds)))
        value_range = max(df["upper"]) - min(df["lower"])

        tooltip_parts.append((f"Group {group}", "@y"))

        range_max = max(df["upper"]) + value_range * 0.1
        if reward_threshold is not None:
            range_max = max(range_max, reward_threshold * 1.1)

        # ERROR BAND
        error_band = bokeh.models.Band(
            base="x", lower="lower", upper="upper",
            source=ColumnDataSource(df.reset_index()),
            fill_color=palette[i % len(palette)],
            fill_alpha=0.2,
            line_color=palette[i % len(palette)],
            line_alpha=0.4,
        )
        p.add_layout(error_band)
        p.renderers.extend([error_band])

        # REWARD LINE
        reward_line = p.line(x, means, line_width=2, color=palette[i % len(palette)])
        names_and_lines.append((
            str(group), [reward_line]
        ))

        i += 1

    # tooltip_parts.append(("Cycle", "@x"))
    # tooltip = HoverTool(
    #     tooltips=tooltip_parts,
    #     mode="vline"
    # )
    #
    # p.add_tools(tooltip)

    p.add_tools(bokeh.models.BoxZoomTool())
    p.add_tools(bokeh.models.ResetTool())
    p.add_tools(bokeh.models.SaveTool())

    # REWARD THRESHOLD
    if reward_threshold is not None:
        p.line(x_all, [reward_threshold for _ in x_all], line_color="green", line_width=2, line_alpha=0.7, line_dash="dashed",
               legend_label="Solution Threshold")

    p.y_range = bokeh.models.Range1d(y_min, y_max)

    legend = bokeh.models.Legend(items=names_and_lines, location="bottom_center", orientation="horizontal")
    p.add_layout(legend, "below")

    style_plot(p)
    p.height = 750

    return embed.components(p)


def plot_length_progress(lengths: Dict[str, list], cycles_loaded):
    """Plot development of the lengths of episodes throughout training."""
    means, stds = lengths["mean"], lengths["stdev"]
    stds = np.array(stds)
    stds[stds == None] = 0
    stds = stds * 0.2
    means = np.array(means)
    means[means == None] = 0

    x = list(range(len(means)))
    df = pd.DataFrame(data=dict(x=x, y=means, lower=np.subtract(means, stds), upper=np.add(means, stds)))
    value_range = max(df["upper"]) - min(df["lower"])

    tooltip = HoverTool(
        tooltips=[("Cycle", "@x"),
                  ("Length", "@y")],
        mode="vline"
    )

    p = figure(title="Average Length of Episodes",
               x_axis_label='Cycle',
               y_axis_label='Total Episode Return',
               y_range=(min(df["lower"]), max(df["upper"]) + value_range * 0.1),
               x_range=(0, max(x)),
               **plot_styling)

    p.add_tools(tooltip)
    p.add_tools(bokeh.models.BoxZoomTool())
    p.add_tools(bokeh.models.ResetTool())
    p.add_tools(bokeh.models.SaveTool())

    # ERROR BAND
    error_band = bokeh.models.Band(
        base="x", lower="lower", upper="upper",
        source=ColumnDataSource(df.reset_index()),
        fill_color=palette[1],
        fill_alpha=0.2,
        line_color=palette[1],
        line_alpha=0.4,
    )
    p.add_layout(error_band)
    p.renderers.extend([error_band])

    # REWARD LINE
    p.line(x, means, legend_label="Episode Length (time steps)", line_width=2, color=palette[1])

    # MAX VALUE MARKING
    x_max = np.argmax(means)
    y_max = np.max(means)
    p.add_layout(bokeh.models.Arrow(end=bokeh.models.NormalHead(size=15,
                                                                line_color="darkred",
                                                                line_width=2,
                                                                fill_color="red"),
                                    line_color="darkred",
                                    line_width=2,
                                    x_start=x_max, y_start=y_max + value_range * 0.1,
                                    x_end=x_max, y_end=y_max))
    p.add_layout(bokeh.models.Label(x=x_max, y=y_max + value_range * 0.1, text=str(y_max),
                                    border_line_color='black', border_line_alpha=1.0,
                                    background_fill_color='white', background_fill_alpha=1.0, text_align="center",
                                    text_line_height=1.5, text_font_size="10pt",

                                    ))

    load_markings = []
    for load_timing in cycles_loaded:
        load_markings.append(
            Span(location=load_timing[0], dimension="height", line_color="red", line_width=2, line_dash=[6, 3])
        )

    p.renderers.extend(load_markings)

    p.legend.location = "bottom_right"
    style_plot(p)

    return embed.components(p)


def plot_distribution(metric: List[float], name="Metric", color=0) -> Tuple:
    hist, edges = np.histogram(metric, density=True, bins=50)

    p = figure(title=f"Distribution of {name}",
               x_axis_label=name,
               y_range=(0, np.max(hist)),
               **plot_styling)

    width = (max(metric) - min(metric)) / len(hist) * 0.9
    hist_glyph = p.rect(x=edges, y=hist / 2, height=hist, width=width, line_color="white", fill_color=palette[color])

    x, mean, sigma = np.linspace(min(metric), max(metric), 50), np.mean(metric), np.std(metric)
    pdf = stats.gaussian_kde(metric, bw_method=0.4).pdf(x)
    p.line(x, pdf, line_color=palette[color], line_width=4, alpha=0.7, )

    tooltip = HoverTool(
        tooltips=[("Reward", "@x"),
                  ("Density", "@y")],
        mode="vline",
        renderers=[hist_glyph]
    )
    p.add_tools(tooltip)
    p.add_tools(bokeh.models.SaveTool())

    p.legend.location = "bottom_right"
    style_plot(p)

    return embed.components(p)


def plot_loss(loss, rewards, name, color_id=0):
    """Plot a loss as it develops over cycles."""
    x = list(range(len(loss)))
    loss = np.array(loss, dtype=float)
    rewards = np.array(rewards, dtype=float)

    loss[np.isnan(loss)] = None

    p = figure(title=name,
               x_axis_label='Cycle',
               y_axis_label='Loss',
               y_range=(min(loss), max(loss)),
               **plot_styling)

    p.extra_y_ranges = {"Reward": Range1d(start=min(rewards), end=max(rewards))}
    p.add_layout(LinearAxis(y_range_name="Reward"), "right")

    p.line(x, rewards, legend_label="Reward", line_width=2, color="lightgrey", y_range_name="Reward")
    p.line(x, loss, legend_label=name, line_width=2, color=palette[color_id])

    p.add_tools(bokeh.models.SaveTool())


    p.legend.location = "bottom_right"
    style_plot(p)

    return embed.components(p)


def plot_preprocessor(preprocessor_data: Dict[str, List[Dict[str, float]]]):
    """Plot progression of preprocessor means."""
    x = list(range(len(preprocessor_data["mean"])))

    plots = []
    for i, sense in enumerate(preprocessor_data["mean"][0].keys()):
        p = figure(title=f"{sense.capitalize()}",
                   x_axis_label='Cycle',
                   y_axis_label='Running Mean',
                   **plot_styling)

        trace = [point[sense] for point in preprocessor_data["mean"]]
        if len(trace) > 1:
            trace = trace[1:]

            if i == 0:
                x = x[1:]

        p.line(x, trace, legend_label=str(sense), line_width=2, color=palette[i])

        p.add_tools(bokeh.models.SaveTool())

        p.legend.location = "bottom_right"
        style_plot(p)

        plots.append(p)

    p = bokeh.layouts.row(plots)
    p.sizing_mode = "stretch_width"
    return embed.components(p)



from angorapy.common.transformers import StateNormalizationTransformer
from angorapy.common.wrappers import make_env

if __name__ == '__main__':
    import matplotlib.pyplot as plt

    env = make_env("OpenAIManipulateApproxDiscrete-v0", transformers=[StateNormalizationTransformer])

    state = env.reset()
    env.render()
    done = False
    for j in range(10000):
        o, r, d, i = env.step(env.action_space.sample())
        if j % 10 == 0:
            env.render()





"""Module for additional environments as well as registering modified environments."""
import os

import gym
from mpi4py import MPI

from angorapy.common.const import SHADOWHAND_MAX_STEPS, SHADOWHAND_SEQUENCE_MAX_STEPS, N_SUBSTEPS

try:
    import mujoco
    has_mujoco = True

    if MPI.COMM_WORLD.rank == 0:
        print("A MuJoCo path exists. MuJoCo is being loaded...")

except ImportError as e:
    has_mujoco = False

    if MPI.COMM_WORLD.rank == 0:
        print("No MuJoCo path exists. MuJoCo is not going to be loaded...")

if has_mujoco:
    from angorapy.environments.adapted import InvertedPendulumNoVelEnv, ReacherNoVelEnv, HalfCheetahNoVelEnv, \
        LunarLanderContinuousNoVel, LunarLanderMultiDiscrete
    from angorapy.environments.manipulate import ManipulateBlock, ManipulateEgg, ManipulateBlockDiscrete, OpenAIManipulate, \
        OpenAIManipulateDiscrete, HumanoidManipulateBlockDiscrete, HumanoidManipulateBlockDiscreteAsynchronous, \
        HumanoidManipulateBlockAsynchronous, HumanoidManipulateBlock
    from angorapy.environments.reach import Reach, FreeReach, FreeReachSequential, ReachSequential

    # SHADOW HAND
    gym.envs.register(
        id='BaseShadowHandEnv-v0',
        entry_point='angorapy.environments:ManipulateBlock',
        kwargs={"visual_input": True, "max_steps": SHADOWHAND_MAX_STEPS},
    )

    gym.envs.register(
        id='ShadowHandBlind-v0',
        entry_point='angorapy.environments:ManipulateBlock',
        kwargs={"visual_input": False, "max_steps": SHADOWHAND_MAX_STEPS},
    )

    # REACHING

    gym.envs.register(
        id='HandReachDenseAbsolute-v1',
        entry_point='angorapy.environments:OldShadowHandReach',
        kwargs={"reward_type": "dense", "relative_control": False, "success_multiplier": 0.1},
        max_episode_steps=100,
    )

    gym.envs.register(
        id='HandReachDenseAbsoluteFine-v1',
        entry_point='angorapy.environments:OldShadowHandReach',
        kwargs={"reward_type": "dense", "n_substeps": 1, "relative_control": False, "success_multiplier": 0.1},
        max_episode_steps=100,
    )

    for step_granularity in ["Fine", ""]:
        for vision_mode in ["Visual", ""]:
            for control_mode in ["Relative", "Absolute"]:
                for init_mode in ["", "Random", "Buffered"]:
                    gym.envs.register(
                        id=f'Reach{init_mode}{control_mode}{vision_mode}{step_granularity}-v0',
                        entry_point='angorapy.environments:Reach',
                        kwargs={"relative_control": control_mode == "Relative", "vision": vision_mode == "Visual",
                                **({"initial_qpos": init_mode.lower()} if init_mode else {}),
                                "n_substeps": 1 if step_granularity == "Fine" else N_SUBSTEPS},
                        max_episode_steps=SHADOWHAND_MAX_STEPS,
                    )

                    gym.envs.register(
                        id=f'Reach{init_mode}{control_mode}{vision_mode}NoTouch{step_granularity}-v0',
                        entry_point='angorapy.environments:Reach',
                        kwargs={"relative_control": control_mode == "Relative", "vision": vision_mode == "Visual",
                                **({"initial_qpos": init_mode.lower()} if init_mode else {}), "touch": False,
                                "n_substeps": 1 if step_granularity == "Fine" else N_SUBSTEPS},
                        max_episode_steps=SHADOWHAND_MAX_STEPS,
                    )

                    gym.envs.register(
                        id=f'FreeReach{init_mode}{control_mode}{vision_mode}{step_granularity}-v0',
                        entry_point='angorapy.environments:FreeReach',
                        kwargs={"relative_control": control_mode == "Relative", "vision": vision_mode == "Visual",
                                **({"initial_qpos": init_mode.lower()} if init_mode else {}),
                                "n_substeps": 1 if step_granularity == "Fine" else N_SUBSTEPS},
                        max_episode_steps=SHADOWHAND_MAX_STEPS,
                    )

                    for i, name in enumerate(["FF", "MF", "RF", "LF"]):
                        gym.envs.register(
                            id=f'FreeReach{name}{init_mode}{control_mode}{vision_mode}{step_granularity}-v0',
                            entry_point='angorapy.environments:FreeReach',
                            kwargs={"relative_control": control_mode == "Relative",
                                    "vision": vision_mode == "Visual",
                                    "force_finger": i,
                                    **({"initial_qpos": init_mode.lower()} if init_mode else {}),
                                    "n_substeps": 1 if step_granularity == "Fine" else N_SUBSTEPS},
                            max_episode_steps=SHADOWHAND_MAX_STEPS,
                        )

                        gym.envs.register(
                            id=f'Reach{name}{init_mode}{control_mode}{vision_mode}{step_granularity}-v0',
                            entry_point='angorapy.environments:Reach',
                            kwargs={"relative_control": control_mode == "Relative",
                                    "vision": vision_mode == "Visual",
                                    "force_finger": i,
                                    **({"initial_qpos": init_mode.lower()} if init_mode else {}),
                                    "n_substeps": 1 if step_granularity == "Fine" else N_SUBSTEPS},
                            max_episode_steps=SHADOWHAND_MAX_STEPS,
                        )

                    # REACH SEQUENCES
                    gym.envs.register(
                        id=f'FreeReachSequential{init_mode}{control_mode}{vision_mode}{step_granularity}-v0',
                        entry_point='angorapy.environments:FreeReachSequential',
                        kwargs={"relative_control": control_mode == "Relative", "vision": vision_mode == "Visual",
                                **({"initial_qpos": init_mode.lower()} if init_mode else {}),
                                "n_substeps": 1 if step_granularity == "Fine" else N_SUBSTEPS},
                        max_episode_steps=SHADOWHAND_SEQUENCE_MAX_STEPS,
                    )

                    gym.envs.register(
                        id=f'ReachSequential{init_mode}{control_mode}{vision_mode}{step_granularity}-v0',
                        entry_point='angorapy.environments:ReachSequential',
                        kwargs={"relative_control": control_mode == "Relative", "vision": vision_mode == "Visual",
                                **({"initial_qpos": init_mode.lower()} if init_mode else {}),
                                "n_substeps": 1 if step_granularity == "Fine" else N_SUBSTEPS},
                        max_episode_steps=SHADOWHAND_SEQUENCE_MAX_STEPS,
                    )

    # MANIPULATE
    for control_mode in ["Relative", "Absolute"]:
        gym.envs.register(
            id=f'ManipulateBlock{control_mode}-v0',
            entry_point='angorapy.environments:ManipulateBlock',
            kwargs={'target_position': 'ignore', 'target_rotation': 'xyz', "relative_control": control_mode == "Relative"},
            max_episode_steps=SHADOWHAND_MAX_STEPS,
        )

        gym.envs.register(
            id=f'ManipulateBlockDiscrete{control_mode}-v0',
            entry_point='angorapy.environments:ManipulateBlockDiscrete',
            kwargs={'target_position': 'ignore',
                    'target_rotation': 'xyz',
                    "relative_control": control_mode == "Relative"},
            max_episode_steps=SHADOWHAND_MAX_STEPS,
        )

        gym.envs.register(
            id=f'ManipulateEgg{control_mode}-v0',
            entry_point='angorapy.environments:ManipulateEgg',
            kwargs={'target_position': 'ignore', 'target_rotation': 'xyz', "relative_control": control_mode == "Relative"},
            max_episode_steps=SHADOWHAND_MAX_STEPS,
        )

    gym.envs.register(
        id=f'OpenAIManipulate-v0',
        entry_point='angorapy.environments:OpenAIManipulate',
        kwargs={},
        max_episode_steps=50 * 100,
    )

    gym.envs.register(
        id=f'OpenAIManipulateDiscrete-v0',
        entry_point='angorapy.environments:OpenAIManipulateDiscrete',
        kwargs={},
        max_episode_steps=50 * 100,
    )

    gym.envs.register(
        id=f'OpenAIManipulateApproxDiscrete-v0',
        entry_point='angorapy.environments:OpenAIManipulateDiscrete',
        kwargs={"delta_t": 0.008},
        max_episode_steps=50 * 100,
    )

    gym.envs.register(
        id=f'OpenAIManipulateApprox-v0',
        entry_point='angorapy.environments:OpenAIManipulate',
        kwargs={"delta_t": 0.008},
        max_episode_steps=50 * 100,
    )

    gym.envs.register(
        id=f'ManipulateBlockApproxDiscrete-v0',
        entry_point='angorapy.environments:ManipulateBlockDiscrete',
        kwargs={"delta_t": 0.008},
        max_episode_steps=50 * 100,
    )

    gym.envs.register(
        id=f'HumanoidManipulateBlockDiscrete-v0',
        entry_point='angorapy.environments:HumanoidManipulateBlockDiscrete',
        kwargs={"delta_t": 0.008},
        max_episode_steps=50 * 100,
    )

    gym.envs.register(
        id=f'HumanoidManipulateBlockDiscreteAsynchronous-v0',
        entry_point='angorapy.environments:HumanoidManipulateBlockDiscreteAsynchronous',
        kwargs={"delta_t": 0.008},
        max_episode_steps=50 * 100,
    )

    gym.envs.register(
        id=f'HumanoidManipulateBlockAsynchronous-v0',
        entry_point='angorapy.environments:HumanoidManipulateBlockAsynchronous',
        kwargs={"delta_t": 0.008},
        max_episode_steps=50 * 100,
    )

    gym.envs.register(
        id=f'HumanoidManipulateBlock-v0',
        entry_point='angorapy.environments:HumanoidManipulateBlock',
        kwargs={"delta_t": 0.008},
        max_episode_steps=50 * 100,
    )


    # MODIFIED ENVIRONMENTS

    gym.envs.register(
        id='MountainCarLong-v0',
        entry_point='gym.envs.classic_control:MountainCarEnv',
        max_episode_steps=500,
        reward_threshold=-110.0,
    )

    gym.envs.register(
        id="InvertedPendulumNoVel-v2",
        entry_point="angorapy.environments:InvertedPendulumNoVelEnv",
        max_episode_steps=1000,
        reward_threshold=950.0,
    )

    gym.envs.register(
        id='ReacherNoVel-v2',
        entry_point='angorapy.environments:ReacherNoVelEnv',
        max_episode_steps=50,
        reward_threshold=-3.75,
    )

    gym.envs.register(
        id='HalfCheetahNoVel-v2',
        entry_point='angorapy.environments:HalfCheetahNoVelEnv',
        max_episode_steps=1000,
        reward_threshold=4800.0,
    )

    gym.envs.register(
        id='LunarLanderContinuousNoVel-v2',
        entry_point='angorapy.environments:LunarLanderContinuousNoVel',
        max_episode_steps=1000,
        reward_threshold=200,
    )

    gym.envs.register(
        id='LunarLanderMultiDiscrete-v2',
        entry_point='angorapy.environments:LunarLanderMultiDiscrete',
        max_episode_steps=1000,
        reward_threshold=200,
    )



#!/usr/bin/env python
import os
from time import sleep
from typing import List, Union

import gym
import tensorflow as tf

from angorapy.common.policies import BasePolicyDistribution
from angorapy.agent.ppo_agent import PPOAgent
from angorapy.utilities.model_utils import is_recurrent_model, list_layer_names, get_layers_by_names, build_sub_model_to, \
    extract_layers, CONVOLUTION_BASE_CLASS, is_conv
from angorapy.utilities.util import add_state_dims, flatten, insert_unknown_shape_dimensions


class Investigator:
    """Interface for investigating a policy network.

    The class serves as a wrapper to use a collection of methods that can extract information about the network."""

    def __init__(self, network: tf.keras.Model, distribution: BasePolicyDistribution):
        """Build an investigator for a network using a distribution.

        Args:
            network (tf.keras.Model):               the policy network to investigate
            distribution (BasePolicyDistribution):     a distribution that the network predicts
        """
        self.network: tf.keras.Model = network
        self.distribution: BasePolicyDistribution = distribution

    @staticmethod
    def from_agent(agent: PPOAgent):
        """Instantiate an investigator from an agent object."""
        agent.policy, agent.value, agent.joint = agent.build_models(agent.joint.get_weights(),
                                                                    batch_size=1, sequence_length=1)
        return Investigator(agent.policy, agent.distribution)

    def list_layer_names(self, only_para_layers=True) -> List[str]:
        """Get a list of unique string representations of all layers in the network."""
        return list_layer_names(self.network, only_para_layers=only_para_layers)

    def list_convolutional_layer_names(self) -> List[str]:
        """Get a list of unique string representations of convolutional layers in the network."""
        return [layer.name for layer in extract_layers(self.network) if is_conv(layer)]

    def list_non_convolutional_layer_names(self) -> List[str]:
        """Get a list of unique string representations of non-convolutional layers in the network."""
        return [layer.name for layer in extract_layers(self.network) if
                not isinstance(layer, CONVOLUTION_BASE_CLASS) and not isinstance(layer, tf.keras.layers.Activation)]

    def get_layers_by_names(self, layer_names: Union[List[str], str]):
        """Retrieve the layer object identified from the model by its unique string representation."""
        if not isinstance(layer_names, list):
            layer_names = [layer_names]
        return get_layers_by_names(self.network, layer_names=layer_names)

    def get_layer_by_name(self, layer_name):
        """Retrieve the layer object identified from the model by its unique string representation."""
        return self.get_layers_by_names(layer_name)[0]

    def get_layer_weights(self, layer_name):
        """Get the weights of a layer identified by its unique name."""
        return self.get_layers_by_names(layer_name)[0].get_weights()

    def get_weight_dict(self):
        """Get a dictionary mapping layer names to the weights of the layers."""
        out = {}
        for layer_name in self.list_layer_names(only_para_layers=True):
            out[layer_name] = self.get_layer_weights(layer_name)

        return out

    def plot_model(self):
        """Plot the network graph into a file."""
        tf.keras.utils.plot_model(self.network, show_shapes=True)

    def dissect_recurrent_layer_weights(self, layer_name):
        """Return a recurrent cells weights and biases in a named dictionary."""
        layer = self.get_layers_by_names(layer_name)[0]

        if not is_recurrent_model(layer):
            raise ValueError("Cannot dissect non-recurrent layer...")

        units = layer.units

        # stolen from https://stackoverflow.com/a/51484524/5407682
        W = layer.get_weights()[0]
        U = layer.get_weights()[1]
        b = layer.get_weights()[2]

        if isinstance(layer, tf.keras.layers.SimpleRNN):
            return dict(
                W=W,
                U=U,
                b=b
            )
        elif isinstance(layer, tf.keras.layers.GRU):
            return dict(
                W_z=W[:, :units],
                W_r=W[:, units: units * 2],
                W_c=W[:, units * 2:],

                U_z=U[:, :units],
                U_r=U[:, units: units * 2],
                U_c=U[:, units * 2:],

                b_z=b[:units],
                b_r=b[units: units * 2],
                b_c=b[units * 2:],
            )
        elif isinstance(layer, tf.keras.layers.LSTM):
            return dict(
                W_i=W[:, :units],
                W_f=W[:, units: units * 2],
                W_c=W[:, units * 2: units * 3],
                W_o=W[:, units * 3:],

                U_i=U[:, :units],
                U_f=U[:, units: units * 2],
                U_c=U[:, units * 2: units * 3],
                U_o=U[:, units * 3:],

                b_i=b[:units],
                b_f=b[units: units * 2],
                b_c=b[units * 2: units * 3],
                b_o=b[units * 3:],
            )
        else:
            raise ValueError("Recurrent layer type not understood. Is it custom?")

    def get_layer_activations(self, layer_name: str, input_tensor=None):
        """Get activations of a layer. If no input tensor is given, a random tensor is used."""

        # make a sub model to the requested layer
        sub_model = build_sub_model_to(self.network, [layer_name])

        if input_tensor is None:
            input_tensor = tf.random.normal(insert_unknown_shape_dimensions(sub_model.input_shape))

        return sub_model.predict(input_tensor)

    def get_activations_over_episode(self, layer_names: Union[List[str], str], env: gym.Env, render: bool = False):
        """Run an episode using the network and get (serialization, activation, r) tuples for each timestep."""
        layer_names = layer_names if isinstance(layer_names, list) else [layer_names]

        states, activations, reward_trajectory, action_trajectory = [], [], [], []

        # make new model with multiple outputs
        polymodel = build_sub_model_to(self.network, layer_names, include_original=True)
        is_recurrent = is_recurrent_model(self.network)

        done = False
        state = env.reset()
        env.goal = 1
        env.render() if render else ""
        while not done:
            dual_out = flatten(polymodel.predict(add_state_dims(parse_state(state), dims=2 if is_recurrent else 1)))
            activation, probabilities = dual_out[:-len(self.network.output)], dual_out[-len(self.network.output):]

            states.append(state)
            activations.append(activation)

            action = self.distribution.act_deterministic(*probabilities)
            action_trajectory.append(action)
            observation, reward, done, info = env.step(action)
            observation, reward, done, info = self.preprocessor.modulate((parse_state(observation), reward, done, info),
                                                                      update=False)

            state = observation
            reward_trajectory.append(reward)

            env.render() if render else ""

        return [states, list(zip(*activations)), reward_trajectory, action_trajectory]

    def render_episode(self, env: gym.Env, slow_down: bool = False, to_gif: bool = False, substeps_per_step=1,
                       act_confidently=True) -> None:
        """Render an episode in the given environment."""
        is_recurrent = is_recurrent_model(self.network)
        self.network.reset_states()

        done, step = False, 0
        state = env.reset()
        cumulative_reward = 0
        # env.render() if not to_gif else env.render(mode="rgb_array")
        while not done:
            step += 1

            prepared_state = state.with_leading_dims(time=is_recurrent).dict()
            probabilities = flatten(self.network(prepared_state, training=False))

            if act_confidently:
                action = self.distribution.act_deterministic(*probabilities)
            else:
                action, _ = self.distribution.act(*probabilities)

            observation, reward, done, info = env.step(action)
            cumulative_reward += (reward if "original_reward" not in info else info["original_reward"])

            state = observation

            if slow_down:
                sleep(0.1)

            env.render() if not to_gif else env.render(mode="rgb_array")

        print(f"Finished after {step} steps with a score of {round(cumulative_reward, 4)}. "
              f"{'Good Boy!' if env.spec.reward_threshold is not None and cumulative_reward > env.spec.reward_threshold else ''}")


if __name__ == "__main__":
    print("INVESTIGATING")
    os.chdir("../../")
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'

    # agent_id = 1585500821  # cartpole-v1
    agent_id = 1583256614 # reach task
    agent_007 = PPOAgent.from_agent_state(agent_id, from_iteration="b")

    inv = Investigator.from_agent(agent_007)
    print(inv.list_layer_names())

    # inv.get_activations_over_episode("policy_recurrent_layer", agent_007.env)

    # inv.get_activations_over_episode("policy_recurrent_layer", agent_007.env)

    for i in range(100):
        inv.render_episode(agent_007.env, to_gif=False)



"""Wrappers encapsulating environments to modulate n_steps, rewards, and control state initialization."""
import abc
from pprint import pprint
from typing import Union, List, Type, OrderedDict

import gym
import numpy
from mpi4py import MPI

from angorapy.common.senses import Sensation
from angorapy.common.transformers import BaseTransformer, merge_transformers
from angorapy.utilities.util import env_extract_dims


class BaseWrapper(gym.ObservationWrapper, abc.ABC):
    """Abstract base class for preprocessors."""

    def __init__(self, env):
        super().__init__(env)

    @property
    def name(self):
        """The name of the transformers."""
        return self.__class__.__name__

    @abc.abstractmethod
    def warmup(self, n_steps=10):
        """Warmup the environment."""
        pass

    def observation(self, observation):
        """Process an observation to be of type 'Sensation'."""
        if isinstance(observation, Sensation):
            return observation
        elif isinstance(observation, (dict, OrderedDict)):
            assert "observation" in observation.keys(), "Unknown dict type of state couldnt be resolved to Sensation."

            if isinstance(observation["observation"], Sensation):
                return observation["observation"]
            elif isinstance(observation["observation"], numpy.ndarray):  # GOAl ENVS
                return Sensation(proprioception=observation["observation"], goal=observation["desired_goal"])
            elif isinstance(observation["observation"], dict) and all([k in observation["observation"].keys() for k in Sensation.sense_names]):
                return Sensation(**observation["observation"])

        return Sensation(proprioception=observation)

    # SYNCHRONIZATION

    def mpi_sync(self):
        """Synchronize the wrapper and all wrapped wrappers over all MPI ranks."""
        if isinstance(self.env, BaseWrapper):
            self.env.mpi_sync()

        self._mpi_sync()

    @abc.abstractmethod
    def _mpi_sync(self):
        pass

    # SERIALIZATION

    @abc.abstractmethod
    def serialize(self):
        """Serialize the wrappers defining data."""
        pass


class TransformationWrapper(BaseWrapper):
    """Wrapper transforming rewards and observation based on running means."""

    def __init__(self, env, transformers: List[BaseTransformer]):
        super().__init__(env)

        # TODO maybe change this to expect list of types to build itself?
        self.transformers = transformers

    def __contains__(self, item):
        return item in self.transformers

    def step(self, action):
        """PErform a step and transform the results."""
        step_tuple = super().step(action)
        # include original reward in info
        step_tuple[-1]["original_reward"] = step_tuple[1]

        if len(self.transformers) != 0:
            for transformer in self.transformers:
                step_tuple = transformer.transform(step_tuple, update=True)

        return step_tuple

    def add_transformers(self, transformers):
        """Add a list of transformers to the environment."""
        self.transformers.extend(transformers)

    def clear_transformers(self):
        """Clear the list of transformers."""
        self.transformers = []

    def warmup(self, n_steps=10):
        """Warmup the transformers."""
        for t in self.transformers:
            t.warmup(self, n_steps=n_steps)

    def _mpi_sync(self):
        """Synchronise the transformers of the wrapper over all MPI ranks."""
        synced_transformers = []
        for transformer in self.transformers:
            collection = MPI.COMM_WORLD.gather(transformer, root=0)

            if MPI.COMM_WORLD.Get_rank() == 0:
                synced_transformers.append(merge_transformers(collection))

        self.transformers = MPI.COMM_WORLD.bcast(synced_transformers, root=0)

    def serialize(self):
        """Return separate transformer serializations in a list"""
        return [t.serialize() for t in self.transformers]


def make_env(env_name,
             reward_config: Union[str, dict] = None,
             transformers: List[Union[Type[BaseTransformer], BaseTransformer]] = None,
             **kwargs) -> BaseWrapper:
    """Make environment, including a possible reward config and transformers."""
    base_env = gym.make(env_name, **kwargs)
    state_dim, n_actions = env_extract_dims(base_env)

    if transformers is None:
        transformers = []
    elif all(isinstance(t, BaseTransformer) for t in transformers):
        transformers = transformers
    elif all(callable(t) for t in transformers):
        transformers = [t(env_name, state_dim, n_actions) for t in transformers]

    env = TransformationWrapper(base_env, transformers=transformers)

    if reward_config is not None and hasattr(env, "reward_config"):
        env.set_reward_function(reward_config)
        env.set_reward_config(reward_config)

    return env



#!/usr/bin/env python
"""Custom data types for simplifying data communication."""
import itertools
import statistics
from collections import namedtuple
from typing import List, Union

import numpy as np
from mpi4py import MPI

StatBundle = namedtuple("StatBundle", ["numb_completed_episodes", "numb_processed_frames",
                                       "episode_rewards", "episode_lengths", "tbptt_underflow",
                                       "per_receptor_mean"])


def condense_stats(stat_bundles: List[StatBundle]) -> Union[StatBundle, None]:
    """Infer a single StatBundle from a list of StatBundles."""
    return StatBundle(
        numb_completed_episodes=sum([s.numb_completed_episodes for s in stat_bundles]),
        numb_processed_frames=sum([s.numb_processed_frames for s in stat_bundles]),
        episode_rewards=list(itertools.chain(*[s.episode_rewards for s in stat_bundles])),
        episode_lengths=list(itertools.chain(*[s.episode_lengths for s in stat_bundles])),
        tbptt_underflow=round(statistics.mean(map(lambda x: x.tbptt_underflow, stat_bundles)), 2) if (
                stat_bundles[0].tbptt_underflow is not None) else None,
        per_receptor_mean={
            sense: np.mean([s.per_receptor_mean[sense] for s in stat_bundles], axis=0)
            for sense in stat_bundles[0].per_receptor_mean.keys()
        }
    )


def mpi_condense_stats(stat_bundles: List[StatBundle]) -> Union[StatBundle, None]:
    """Pull together the StatBundle lists from the buffer on all workers and infer a single StatBundle."""
    stat_bundles = MPI.COMM_WORLD.gather(stat_bundles, root=0)

    if MPI.COMM_WORLD.Get_rank() == 0:
        stat_bundles = list(itertools.chain(*stat_bundles))
        return condense_stats(stat_bundles)
    else:
        return None



from setuptools import setup, find_packages

setup(
    name='angorapy',
    version='0.7.1',
    description='ANthropomorphic Goal-ORiented Modeling, Learning and Analysis for Neuroscience',
    url='https://github.com/ccnmaastricht/dexterous-robot-hand',
    author='Tonio Weidler',
    author_email='research@tonioweidler.de',
    license='GPL-3.0',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "numpy==1.19.5",
        "Box2D",
        "gym==0.24.0",
        "mujoco",
        "tensorflow==2.4.2",
        "mpi4py==3.1.3",
        "tqdm",
        "simplejson",
        "psutil",
        "scipy",
        "sklearn",
        "argcomplete",
        "matplotlib",
        "scikit-learn==0.24.1",
        "pandas==1.0.4",
        "nvidia-ml-py3",
        "seaborn",
        "distance",
        "protobuf==3.19.0",

        # webinterface
        "itsdangerous==2.0.1",
        "werkzeug==2.0.3",
        "Flask~=1.1.2",
        "Jinja2==3.0.0",
        "bokeh",
        "flask_jsglue"
    ],

    package_data={
        "angorapy": ["environments/assets/**/*"],
    },

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



"""Helper functions."""
import random
import sys
from typing import Tuple, Union, List, Dict
import os

import gym
import numpy
import numpy as np
import tensorflow as tf
from gym import spaces
from gym.spaces import Discrete, Box, MultiDiscrete
from mpi4py import MPI
from tensorflow.python.client import device_lib

from angorapy.utilities.error import UninterpretableObservationSpace


def get_available_gpus():
    """Get list of available GPUs."""
    local_device_protos = device_lib.list_local_devices()
    return [x.name for x in local_device_protos if x.device_type == 'GPU']


def mpi_flat_print(string: str):
    """A bit of a workaround to no new line printing to have it work in PyCharm."""
    if MPI.COMM_WORLD.Get_rank() == 0:
        print(f"\r{string}", end="")


def set_all_seeds(seed):
    """Set all random seeds (tf, np, random) to given value."""
    tf.random.set_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


def env_extract_dims(env: gym.Env) -> Tuple[Dict[str, Tuple], Tuple[int]]:
    """Returns state and action space dimensionality for given environment."""
    obs_dim: dict

    # extract observation space dimensionalities from environment
    if isinstance(env.observation_space, spaces.Dict):
        # dict observation with observation field, for all GoalEnvs
        if isinstance(env.observation_space["observation"], spaces.Box):
            obs_dim = {"proprioception": env.observation_space["observation"].shape, "goal": env.observation_space["desired_goal"].shape}
        elif isinstance(env.observation_space["observation"], spaces.Dict):
            # made for sensation
            obs_dim = {name: field.shape for name, field in env.observation_space["observation"].spaces.items()}
        else:
            raise UninterpretableObservationSpace(f"Cannot extract the dimensionality from a spaces.Dict observation space "
                                                  f"where the observation is of type "
                                                  f"{type(env.observation_space['observation']).__name__}")
    elif isinstance(env.observation_space, gym.spaces.Box):  # standard observation in box form
        obs_dim = {"proprioception": env.observation_space.shape}
    else:
        raise UninterpretableObservationSpace(
            f"Cannot interpret observation space of type {type(env.observation_space)}.")

    # action space
    if isinstance(env.action_space, Discrete):
        act_dim = (env.action_space.n,)
    elif isinstance(env.action_space, MultiDiscrete):
        assert np.alltrue(env.action_space.nvec == env.action_space.nvec[0]), "Can only handle multi-discrete action" \
                                                                              "spaces where all actions have the same " \
                                                                              "number of categories."
        act_dim = (env.action_space.shape[0], env.action_space.nvec[0].item())
    elif isinstance(env.action_space, Box):
        act_dim = (env.action_space.shape[0], 1)
    else:
        raise NotImplementedError(f"Environment has unknown Action Space Typ: {env.action_space}")

    return obs_dim, act_dim


def normalize(x, is_img=False) -> numpy.ndarray:
    """Normalize a numpy array to have all values in range (0, 1)."""
    x = tf.convert_to_tensor(x).numpy()
    return x / 255 if is_img else (x - x.min()) / (x.max() - x.min())


def flatten(some_list):
    """Flatten a python list."""
    return [some_list] if not isinstance(some_list, list) else [x for X in some_list for x in flatten(X)]


def is_array_collection(a: numpy.ndarray) -> bool:
    """Check if an array is an array of objects (e.g. goal arrays) or an actual array of direct data."""
    return a.dtype == "O"


def add_state_dims(state: "Sensation", dims: int = 1, axis: int = 0) -> 'Sensation':
    """Expand state (array or lost of arrays) to have a batch and/or time dimension."""
    if dims < 1:
        return state

    new_state = state.with_leading_dims(time=dims == 2)

    return new_state


def merge_into_batch(list_of_states: List[Union[numpy.ndarray, Tuple]]):
    """Merge a list of states into one huge batch of states. Handles both single and multi input states.

    Assumes NO batch dimension!
    """
    if isinstance(list_of_states[0], numpy.ndarray):
        return numpy.concatenate(add_state_dims(list_of_states))
    else:
        return tuple(numpy.concatenate(list(map(lambda x: add_state_dims(x[i]), list_of_states)), axis=0)
                     for i in range(len(list_of_states[0])))


def insert_unknown_shape_dimensions(shape, none_replacer: int = 1):
    """Replace Nones in a shape tuple with 1 or a given goal value."""
    return tuple(map(lambda s: none_replacer if s is None else s, shape))


def detect_finished_episodes(dones: tf.Tensor):
    """Detect which samples in the batch connect to a episode that finished during the subsequence, based on the dones
     and return a 1D boolean tensor.

    Input Shape:
        dones: (B, S)
    """
    finished = tf.math.reduce_any(dones, axis=-1)
    return finished


def find_divisors(number: int):
    divisors = []
    for i in range(1, number // 2 + 1):
        if number % i == 0:
            divisors.append(i)
            divisors.append(number // i)

    return list(sorted(set(divisors)))


def find_optimal_tile_shape(floor_shape: Tuple[int, int], tile_size: int) -> Tuple[int, int]:
    """For a given shape of a matrix (floor), find the shape of a tiles that fit the floor and contain
    exactly tile_size elements."""
    height_divisors = list(reversed(find_divisors(floor_shape[0])))
    width_divisors = find_divisors(floor_shape[1])

    for hd in height_divisors:
        for wd in width_divisors:
            if hd * wd == tile_size:
                return hd, wd

    raise ValueError(f"No tiling of size {tile_size} possible for a floor of shape {floor_shape}.")


class HiddenPrints:
    """Context that hides print calls."""

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


if __name__ == "__main__":
    from configs import hp_config

    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '4'
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'

    conf = hp_config.manipulate
    n_optimizers = 32

    total_tiling = find_optimal_tile_shape(
        (conf["workers"], conf["horizon"] // conf["tbptt"]),
        tile_size=conf["batch_size"] // conf["tbptt"]
    )

    optimizer_tiling = find_optimal_tile_shape(
        (conf["workers"] // n_optimizers, conf["horizon"] // conf["tbptt"] // n_optimizers),
        tile_size=conf["batch_size"] // conf["tbptt"] // n_optimizers
    )

    print(f"Total tiling: {total_tiling}\n"
          f"Optimizer's Tiling: {optimizer_tiling}")


from gym import spaces
from gym.envs.box2d import LunarLanderContinuous
from gym.envs.mujoco import HalfCheetahEnv, InvertedPendulumEnv, ReacherEnv
import numpy as np


class InvertedPendulumNoVelEnv(InvertedPendulumEnv):
    """InvertedPendulum Environment without velocities directly encoded in the state."""

    def _get_obs(self):
        return self.sim.data.qpos.ravel()


class ReacherNoVelEnv(ReacherEnv):
    """InvertedPendulum Environment without velocities directly encoded in the state."""

    def _get_obs(self):
        theta = self.sim.data.qpos.flat[:2]
        return np.concatenate([
            np.cos(theta),
            np.sin(theta),
            self.sim.data.qpos.flat[2:],
            # self.sim.data.qvel.flat[:2],
            self.get_body_com("fingertip") - self.get_body_com("target")
        ])


class HalfCheetahNoVelEnv(HalfCheetahEnv):
    """HalfCheetah Environment without velocities directly encoded in the state."""

    def _get_obs(self):
        return self.sim.data.qpos.flat[1:]


class LunarLanderContinuousNoVel(LunarLanderContinuous):
    """LunarLander Continuous Environment without velocities directly encoded in the state."""

    def __init__(self):
        super().__init__()
        self.observation_space = spaces.Box(-np.inf, np.inf, shape=(5,), dtype=np.float32)
        self.reset()

    def step(self, action):
        o, r, d, info = super().step(action)
        o = np.concatenate([o[0:2], [o[4]], o[6:]])

        return o, r, d, info


class LunarLanderMultiDiscrete(LunarLanderContinuous):
    """LunarLander Continuous Environment but again made discrete."""

    discrete_bin_count = 10

    def __init__(self):
        self.discrete_action_values = np.linspace(-1, 1, LunarLanderMultiDiscrete.discrete_bin_count)
        super().__init__()
        self.action_space = spaces.MultiDiscrete(np.ones(2) * LunarLanderMultiDiscrete.discrete_bin_count)

        self.reset()

    def step(self, action: np.ndarray):
        action = self.discrete_action_values[action.astype(np.int)]

        return super().step(action)